from datetime import date

import pytest

from xferx.pdp11.dos11fs import (
    DOS11Filesystem,
    UserFileDirectoryBlock,
    date_to_dos11,
    dos11_to_date,
)
from xferx.shell import Shell

DSK = "tests/dsk/dos11_rk05.dsk"


def test_dos11():
    shell = Shell(verbose=True)
    shell.onecmd(f"mount t: /dos11 {DSK}", batch=True)
    fs = shell.volumes.get('T')
    assert isinstance(fs, DOS11Filesystem)

    shell.onecmd("dir t:", batch=True)
    shell.onecmd("dir /uic t:", batch=True)
    shell.onecmd("type t:1.txt", batch=True)

    x = fs.read_bytes("50.txt")
    x = x.rstrip(b"\0")
    assert len(x) == 2200
    for i in range(0, 50):
        assert f"{i:5d} ABCDEFGHIJKLMNOPQRSTUVWXYZ01234567890".encode("ascii") in x

    l = list(fs.filter_entries_list("*.TXT[100,100]"))
    assert len(l) == 3
    assert all([x.contiguous for x in l])

    x1 = fs.read_bytes("[100,100]1000.txt")
    x1 = x1.rstrip(b"\0")
    assert len(x1) == 44000
    for i in range(0, 1000):
        assert f"{i:5d} ABCDEFGHIJKLMNOPQRSTUVWXYZ01234567890".encode("ascii") in x1

    l = list(fs.filter_entries_list("*.TXT[200,200]"))
    assert len(l) == 3
    assert all([not x.contiguous for x in l])

    x2 = fs.read_bytes("[100,100]1000.txt")
    x2 = x2.rstrip(b"\0")
    assert len(x2) == 44000
    for i in range(0, 1000):
        assert f"{i:5d} ABCDEFGHIJKLMNOPQRSTUVWXYZ01234567890".encode("ascii") in x2
    assert x1 == x2

    l = list(fs.filter_entries_list("*.TXT[1,2]"))
    assert len(l) == 0

    x3 = fs.read_bytes("[200,200]500.txt")
    x3 = x3.rstrip(b"\0")
    assert len(x3) == 22000
    for i in range(0, 500):
        assert f"{i:5d} ABCDEFGHIJKLMNOPQRSTUVWXYZ01234567890".encode("ascii") in x3


def test_dos11_bitmap():
    shell = Shell(verbose=True)
    shell.onecmd(f"copy {DSK} {DSK}.mo", batch=True)
    shell.onecmd(f"mount t: /dos11 {DSK}.mo", batch=True)
    fs = shell.volumes.get('T')
    assert isinstance(fs, DOS11Filesystem)

    d = fs.get_file_entry("[200,200]500.TXT")
    assert d is not None
    assert not d.contiguous

    e = fs.get_file_entry("[100,100]200.TXT")
    assert e is not None
    assert e.contiguous

    # Test is_free
    bitmap = fs.read_bitmap()
    for i in range(e.start_block, e.start_block + e.length):
        assert not bitmap.is_free(i)
    assert bitmap.is_free(187)
    assert bitmap.is_free(4649)

    # Test find_contiguous_blocks
    assert bitmap.find_contiguous_blocks(10) == 4640
    assert bitmap.find_contiguous_blocks(1000) == 3650
    with pytest.raises(OSError):
        bitmap.find_contiguous_blocks(10000)
    assert bitmap.used() == 337
    d_length = d.length
    e_length = e.length

    # Write UFD
    e.ufd_block.write()
    ufd_block2 = UserFileDirectoryBlock.read(e.ufd_block.fs, e.ufd_block.uic, e.ufd_block.block_number)
    assert str(e.ufd_block) == str(ufd_block2)

    # Delete contiguous file
    e.delete()
    with pytest.raises(FileNotFoundError):
        fs.get_file_entry("[100,100]200.TXT")
    bitmap = fs.read_bitmap()
    assert bitmap.used() == 337 - e_length

    # Delete linked file
    d.delete()
    with pytest.raises(FileNotFoundError):
        fs.get_file_entry("[200,200]500.TXT")
    bitmap = fs.read_bitmap()
    assert bitmap.used() == 337 - e_length - d_length

    # UIC not found
    with pytest.raises(Exception):
        shell.onecmd("copy /TYPE:CONTIGUOUS t:10.TXT t:[123,321]10NEW.TXT", batch=True)

    # Create a contiguous file
    shell.onecmd("copy /TYPE:CONTIGUOUS t:10.TXT t:[100,100]10NEW.TXT", batch=True)
    x2 = fs.read_bytes("[100,100]10NEW.txt")
    x2 = x2.rstrip(b"\0")
    assert len(x2) == 440
    for i in range(0, 10):
        assert f"{i:5d} ABCDEFGHIJKLMNOPQRSTUVWXYZ01234567890".encode("ascii") in x2

    for i in range(0, 100):
        shell.onecmd(f"copy /TYPE:CONTIGUOUS t:1.TXT t:[100,100]A{i}.TXT", batch=True)

    # Create a non-contiguous file
    shell.onecmd("copy /TYPE:NOCONTIGUOUS t:10.TXT t:[200,200]10NEW.TXT", batch=True)
    x2 = fs.read_bytes("[200,200]10NEW.txt")
    x2 = x2.rstrip(b"\0")
    assert len(x2) == 440
    for i in range(0, 10):
        assert f"{i:5d} ABCDEFGHIJKLMNOPQRSTUVWXYZ01234567890".encode("ascii") in x2


def test_dos11_to_date():
    assert dos11_to_date(0) is None
    assert dos11_to_date(21163) == date(1991, 6, 12)
    assert dos11_to_date(16134) == date(1986, 5, 14)


def test_date_to_dos11():
    assert date_to_dos11(None) == 0
    assert date_to_dos11(date(1991, 6, 12)) == 21163
    assert date_to_dos11(date(1986, 5, 14)) == 16134


def test_date_combined():
    assert dos11_to_date(date_to_dos11(None)) is None
    original_date = date(1980, 5, 28)
    dos11_date = date_to_dos11(original_date)
    converted_date = dos11_to_date(dos11_date)
    assert original_date == converted_date


def test_dos11_create_uic():
    shell = Shell(verbose=True)
    shell.onecmd(f"copy {DSK} {DSK}.mo", batch=True)
    shell.onecmd(f"mount t: /dos11 {DSK}.mo", batch=True)
    fs = shell.volumes.get('T')
    assert isinstance(fs, DOS11Filesystem)
    # Create the UIC
    shell.onecmd("create /directory t:[10,20]", batch=True)
    with pytest.raises(Exception):
        shell.onecmd("create /directory t:[10,20]", batch=True)
    shell.onecmd("create t:[10,20]test /allocate:5", batch=True)
    fs.get_file_entry("[10,20]test")
    shell.onecmd("dir t:[10,20]", batch=True)
    # Delete the UIC
    shell.onecmd("delete t:[10,20]", batch=True)
    with pytest.raises(FileNotFoundError):
        fs.get_file_entry("[10,20]test")
    with pytest.raises(Exception):
        shell.onecmd("delete t:[10,20]", batch=True)


def test_dos11_init():
    shell = Shell(verbose=True)
    shell.onecmd(f"mount in: /dos11 {DSK}", batch=True)
    shell.onecmd(f"create /allocate:2400 {DSK}.mo", batch=True)
    shell.onecmd(f"init /dos11 {DSK}.mo", batch=True)
    shell.onecmd(f"mount ou: /dos11 {DSK}.mo", batch=True)
    shell.onecmd("dir ou:", batch=True)
    shell.onecmd("copy in:* ou:", batch=True)
    fs = shell.volumes.get('OU')
    assert isinstance(fs, DOS11Filesystem)

    # Create the UIC
    shell.onecmd("create /directory ou:[10,20]", batch=True)
    with pytest.raises(Exception):
        shell.onecmd("create /directory ou:[10,20]", batch=True)
    shell.onecmd("create ou:[10,20]test /allocate:5", batch=True)
    fs.get_file_entry("[10,20]test")
    shell.onecmd("dir ou:[10,20]", batch=True)



# Copyright (C) 2014 Andrea Bonomi <andrea.bonomi@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import errno
import os
import sys
import typing as t
from dataclasses import dataclass
from datetime import date

from ..abstract import AbstractDirectoryEntry, AbstractFile, AbstractFilesystem
from ..commons import (
    ASCII,
    IMAGE,
    READ_FILE_FULL,
    BlockDirection,
    Direction,
    filename_match,
    pairwise,
)
from ..device.abstract import AbstractDevice
from ..device.block_18bit import (
    BlockDevice18Bit,
    from_18bit_words_to_bytes,
    from_bytes_to_18bit_words,
)

__all__ = [
    "ADSSFilesystem",
    "ADSSFile",
    "ADSSDirectoryEntry",
    "ascii_to_five_seven",
    "five_seven_to_ascii",
    "decode_block_format",
    "encode_block_format",
]

# PDP-9 ADVANCED SOFTWARE SYSTEM MONITORS, Pag 95
# https://bitsavers.org/pdf/dec/pdp9/DEC-9A-MAD0-D.pdf

# PDP-15 Advanced Monitor Software System for PDP-1S/20/30/40
# PROGRAMMER'S REFERENCE MANUAL  Pag 80
# https://bitsavers.org/pdf/dec/pdp15/DEC-15-MR2B-D_AdvMonPgmRef.pdf

WORDS_PER_BLOCK = 256  # Number of words per block
DIRECTORY_BLOCK = 0o100  # Directory block number on a DECtape
DIRECTORY_WORDS = 0o400  # Non-system tape
DIRECTORY_SYSTEM_WORD = 0o203  # If this word is set to 0o777777, the tape is a system tape
SYSTEM_TAPE_DIRECTORY_WORDS = 0o200  # System tape
BITMAP_WORDS = 0o40  # Bitmap length in words
DIRECTORY_ENTRY_WORDS = 4  # Directory entry length in words
DIRECTORY_ENTRIES = (WORDS_PER_BLOCK - BITMAP_WORDS) // DIRECTORY_ENTRY_WORDS  # Number of directory entries per block
SYSTEM_TAPE_DIRECTORY_ENTRIES = 24  # Number of directory entries on a system tape
FILE_BITMAP_BLOCK = 0o71  # File bitmap block number
FILE_BITMAP_BLOCKS = 7  # Number of blocks for file bitmaps
BITMAPS_PER_BLOCK = WORDS_PER_BLOCK // BITMAP_WORDS  # Number of bitmaps per block
DECTAPE_BLOCKS = 384  # Number of blocks on a DECtape
COMMAND_TABLE_ENTRY_WORDS = 7  # Command table entry length in words
SYSBLK_SYS = "SYSBLK;SYS"

# Data Modes
# Pag 28, pag 134
# https://bitsavers.org/pdf/dec/pdp15/DEC-15-MR2B-D_AdvMonPgmRef.pdf
# Mode 0 - IOPS Binary
# Mode 1 - Image Binary
# Mode 2 - IOPS ASCII
# Mode 3 - Image Alphanumeric
# Mode 4 - Dump
# Mode 5 - 9-Channe1 Dump

BLOCK_ID_IOPS_BINARY = 0  # IOPS Binary
BLOCK_ID_IOPS_ASCII = 2  # IOPS ASCII
BLOCK_ID_EOF = 5  # End of File

assert DIRECTORY_ENTRIES == 56


@dataclass
class BlockContent:
    block: BlockDirection
    words: t.List[int]


def sixbit_to_ascii(word: int) -> str:
    """
    Convert a 18-bit word to an ASCII string
    """
    tmp = [(word >> x) & 0o77 for x in (12, 6, 0)]
    return ("".join(chr(ch if ch >= 32 else 64 + ch) for ch in tmp)).replace("@", "")


def ascii_to_sixbit(val: str) -> int:
    """
    Convert an ASCII string to a 18-bit word
    """
    if len(val) > 3:
        raise ValueError("String must be 3 characters or less")
    val = val.ljust(3, "@")  # Pad
    word = 0
    pow = (12, 6, 0)
    for i, ch in enumerate(val):
        word += ((ord(ch) - 32 if ord(ch) >= 96 else ord(ch)) & 0o77) << pow[i]
    return word


def get_file_mode(fullname: str) -> str:
    """
    Get the file mode based on the filename extension.
    If the extension is not recognized, return IMAGE.
    """
    _, extension = adss_split_ext(fullname)
    if extension == "SRC" or extension[0:1].isnumeric():
        return ASCII
    else:
        return IMAGE


def five_seven_to_ascii(words: t.List[int]) -> bytes:
    """
    Convert a list of 18-bit words using 5/7 ASCII encoding to a string.
    5/7 ASCII refers to the following encoding scheme:
    Five 7-bit ASCII characters are packed in two contiguous locations.

            0                 6   7               13   14           17
           +--------------------+--------------------+-----------------+
    Word 1 |   1st character    |   2nd character    | 3rd chr bit 1-4 |
           +--------------------+--------------------+-----------------+

            0              2   3                8   10           16  17
           +-----------------+--------------------+-----------------+--+
    Word 2 | 3rd chr bit 5-7 |   4 th character   | 5 th character  |  |
           +-----------------+--------------------+-----------------+--+

    Pag 30
    https://bitsavers.org/pdf/dec/pdp15/DEC-15-MR2B-D_AdvMonPgmRef.pdf
    """
    result = bytearray()
    for word1, word2 in pairwise(words):
        chars = [
            ((word1 >> 11) & 0o177),  # First character
            ((word1 >> 4) & 0o177),  # Second character
            (((word1 & 0o017) << 3) | ((word2 >> 15) & 0o07)),  # Third character
            ((word2 >> 8) & 0o177),  # Fourth character
            ((word2 >> 1) & 0o177),  # Fifth character
        ]
        result.extend(chars)
    return bytes(result)


def ascii_to_five_seven(data: bytes) -> t.List[int]:
    """
    Convert a string to a list of 18-bit words using 5/7 ASCII encoding.
    """
    words = []
    for i in range(0, len(data), 5):
        chars = data[i : i + 5]
        if len(chars) < 5:
            chars += b"\0" * (5 - len(chars))  # Pad with null bytes
        assert len(chars) == 5
        words.append((chars[0] << 11) | (chars[1] << 4) | (chars[2] >> 3))
        words.append(((chars[2] & 0o07) << 15) | (chars[3] << 8) | (chars[4] << 1))
    return words


def decode_block_format(words: t.List[int]) -> bytes:
    """
    Every block recorded includes a two-word Block Control Pair followed by the data.
    The Block Control Pair specifies:
    - the block type (ASCII, binary, EOF, etc.),
    - the length of the block in words (including the Block Control Pair)
    - the checksum of the block

    Pag 86
    https://bitsavers.org/pdf/dec/pdp15/DEC-15-MR2B-D_AdvMonPgmRef.pdf
    """
    result = bytearray()
    position = 0
    while position <= len(words) - 2:
        # Read the Block Control Pair
        block_id = (words[position]) & 0o7  # Block ID
        block_word_counter = words[position] >> 8  # Block Word Count (12 bit)
        # print(f"Block ID: {block_id}, Word Count: {block_word_counter} at position {position}")
        if block_id & BLOCK_ID_EOF or block_word_counter == 0:  # End of File or end of block
            break  # End of the block
        # print(f"Block ID: {block_id}, Word Count: {block_word_counter} at position {position}")
        checksum = words[position + 1]  # Checksum word
        tmp = 0o1000000 - (words[position] + sum(words[position + 2 : position + block_word_counter])) & 0o777777
        if tmp != checksum:
            print(f"Checksum error: expected {tmp}, got {checksum} at position {position}")
        if block_id == BLOCK_ID_IOPS_ASCII:
            # IOPS ASCII format
            data = five_seven_to_ascii(words[position + 2 : position + block_word_counter])
            # Convert carriage return to newline, strip null bytes
            result += data.replace(b"\r", b"\n").rstrip(b"\0")
        elif block_id == BLOCK_ID_IOPS_BINARY:
            # IOPS Binary format
            result += from_18bit_words_to_bytes(words[position + 2 : position + block_word_counter], IMAGE)
        else:
            print(f"Unknown block ID: {block_id} at position {position}")
        position += block_word_counter
    return bytes(result)


def split_data(data: bytes, max_length: int) -> t.List[bytes]:
    """
    Split a byte string into chunks of a maximum length
    """
    return [data[i : i + max_length] for i in range(0, len(data), max_length)]


def encode_block_control_pair(block_id: int, words: t.List[int]) -> t.Tuple[int, int]:
    """
    Create a block control pair for the given block ID and words.
    Returns a tuple of (block_word_counter, checksum).
    """
    block_word_counter = len(words) + 2  # +2 for the block control pair
    word1 = (block_word_counter << 8) | block_id
    checksum = 0o1000000 - (word1 + sum(words)) & 0o777777
    return word1, checksum


def encode_block_format(data: bytes, file_mode: str, words_per_block: int) -> t.Iterator[t.List[int]]:
    """
    Encode data into blocks of 18-bit words according to the specified file mode.
    """
    if file_mode == ASCII:
        words: t.List[int] = []
        split = data.split(b"\n")  # Split lines
        if not split[-1]:
            split = split[:-1]
        for line in split:  # Split lines
            line += b"\r"  # Add carriage return to each line
            for part in split_data(line, 256 - 6):  # TODO check with long lines
                block_id = BLOCK_ID_IOPS_ASCII
                part_words = ascii_to_five_seven(part)
                block_word_counter = len(part_words) + 2
                block_control_pair = encode_block_control_pair(block_id, part_words)
                # Check if the block is full
                if len(words) + block_word_counter + 2 > words_per_block:
                    words += [0, 0]  # End of the block
                    if len(words) < words_per_block:
                        words += [0] * (words_per_block - len(words))  # Pad to the block size
                    yield words
                    words = []
                words += block_control_pair
                words += part_words
        # End of the block
        words += encode_block_control_pair(BLOCK_ID_EOF, [])
        if len(words) < words_per_block:
            words += [0] * (words_per_block - len(words))  # Pad to the block size
        assert len(words) == words_per_block, "Block size mismatch"
        yield words
    else:
        words = []
        for part in split_data(data, 26 * 3):
            block_id = BLOCK_ID_IOPS_BINARY
            part_words = from_bytes_to_18bit_words(part, file_type=IMAGE)
            block_word_counter = len(part_words) + 2
            block_control_pair = encode_block_control_pair(block_id, part_words)
            # Check if the block is full
            if len(words) + block_word_counter + 2 > words_per_block:
                words += [0, 0]  # End of the block
                if len(words) < words_per_block:
                    words += [0] * (words_per_block - len(words))  # Pad to the block size
                yield words
                words = []
            words += block_control_pair
            words += part_words
        # End of the block
        words += encode_block_control_pair(BLOCK_ID_EOF, [])
        if len(words) < words_per_block:
            words += [0] * (words_per_block - len(words))  # Pad to the block size
        assert len(words) == words_per_block, "Block size mismatch"
        yield words


def adss_split_ext(fullname: str) -> t.Tuple[str, str]:
    """
    Split a fullname into the filename and extension.

    Filename and extension are separated by a ";"
    Example: "FILENAM;EXT"
    """
    fullname = fullname.upper()
    try:
        filename, extension = fullname.replace(" ", ";").split(";", 1)
        return filename.strip(), extension.strip()
    except Exception:
        return fullname, ""


def adss_canonical_filename(fullname: str, wildcard: bool = False) -> str:
    """
    Generate the canonical ADSS/DOS-15 filename

    A filename is a string of up to six alphanumeric characters.
    The extension can be up to three characters long.
    Any printing character in the ASCII set can be used
    with the exception of a space, ":", ";", ",", "(" and ")".
    Filename and extension are separated by a ";"
    """
    filename, extension = adss_split_ext(fullname)
    filename = sixbit_to_ascii(ascii_to_sixbit(filename[0:3])) + sixbit_to_ascii(ascii_to_sixbit(filename[3:6]))
    extension = sixbit_to_ascii(ascii_to_sixbit(extension))
    return f"{filename};{extension}"


def oct_dump(words: t.List[int], words_per_line: int = 8) -> None:
    """
    Display contents in octal and sixbit ASCII
    """
    for i in range(0, len(words), words_per_line):
        line = words[i : i + words_per_line]
        ascii = five_seven_to_ascii(line).decode("ascii", errors="ignore")
        printable = "".join(x if x.isprintable() else " " for x in ascii)
        oct_str = " ".join([f"{x:06o}" for x in line])
        sys.stdout.write(f"{i:08o}   {oct_str.ljust(5 * words_per_line)}  {printable}\n")


class ADSSFile(AbstractFile):
    entry: "ADSSDirectoryEntry"
    file_mode: str  # ASCII of IMAGE
    closed: bool

    def __init__(self, entry: "ADSSDirectoryEntry", file_mode: t.Optional[str] = None):
        self.entry = entry
        self.file_mode = file_mode or get_file_mode(entry.basename)
        self.closed = False

    def read_block(
        self,
        block_number: int,
        number_of_blocks: int = 1,
    ) -> bytes:
        """
        Read block(s) of data from the file
        """
        if number_of_blocks == READ_FILE_FULL:
            number_of_blocks = self.entry.get_length()
        if (
            self.closed
            or block_number < 0
            or number_of_blocks < 0
            or block_number + number_of_blocks > self.entry.get_length()
        ):
            raise OSError(errno.EIO, os.strerror(errno.EIO))
        data = bytearray()
        for i, block in enumerate(self.entry.get_block_contents()):
            if i >= block_number:
                data.extend(from_18bit_words_to_bytes(block.words, self.file_mode))
                number_of_blocks -= 1
                if number_of_blocks == 0:
                    break
        return bytes(data)

    def read_words_block(
        self,
        block_number: int,
        number_of_blocks: int = 1,
    ) -> t.List[int]:
        """
        Read block(s) of words from the file
        """
        if number_of_blocks == READ_FILE_FULL:
            number_of_blocks = self.entry.get_length()
        if (
            self.closed
            or block_number < 0
            or number_of_blocks < 0
            or block_number + number_of_blocks > self.entry.get_length()
        ):
            raise OSError(errno.EIO, os.strerror(errno.EIO))
        words: t.List[int] = []
        for i, block in enumerate(self.entry.get_block_contents()):
            if i >= block_number:
                words.extend(block.words)
                number_of_blocks -= 1
                if number_of_blocks == 0:
                    break
        return words

    def write_block(
        self,
        buffer: bytes,
        block_number: int,
        number_of_blocks: int = 1,
    ) -> None:
        """
        Write block(s) of data to the file
        """
        words = from_bytes_to_18bit_words(buffer, self.file_mode)
        self.write_words_block(words, block_number, number_of_blocks)

    def write_words_block(
        self,
        words: t.List[int],
        block_number: int,
        number_of_blocks: int = 1,
    ) -> None:
        """
        Write block(s) of data to the file
        """
        if (
            self.closed
            or block_number < 0
            or number_of_blocks < 0
            or block_number + number_of_blocks > self.entry.get_length()
        ):
            raise OSError(errno.EIO, os.strerror(errno.EIO))
        words = list(words)
        for i, block in enumerate(self.entry.get_block_contents()):
            if i >= block_number:
                block.words[: WORDS_PER_BLOCK - 1] = words[: WORDS_PER_BLOCK - 1]  # Update the block words
                self.entry.fs.write_words_block(block.block, block.words)
                words = words[WORDS_PER_BLOCK - 1 :]  # Remove the written words
                number_of_blocks -= 1
                if number_of_blocks == 0:
                    break

    def get_length(self) -> int:
        """
        Get the length in blocks
        """
        return self.entry.get_length()

    def get_size(self) -> int:
        """
        Get file size in bytes
        """
        return self.get_length() * self.get_block_size()

    def get_block_size(self) -> int:
        """
        Get file block size in bytes
        """
        return (WORDS_PER_BLOCK - 1) * 3

    def close(self) -> None:
        """
        Close the file
        """
        self.closed = True

    def __str__(self) -> str:
        return self.entry.fullname


class ADSSDirectoryEntry(AbstractDirectoryEntry):
    """
    Directory Entry

    Word

        +-----------------------------------+
      0 | File name                         |
      1 |                                   |
        +-----------------------------------+
      2 | Extension                         |
        +-----------------------------------+
      3 | Active bit | Block number         |
        +-----------------------------------+

    """

    directory: "ADSSDirectory"
    file_number: int = 0  # File number in the directory
    filename: str  # Filename
    extension: str = ""  # File extension
    raw_data_link: int = 0  # Raw data link (active bit, block number)

    def __init__(self, directory: "ADSSDirectory"):
        self.directory = directory

    @classmethod
    def new(cls, directory: "ADSSDirectory", file_number: int) -> "ADSSDirectoryEntry":
        """
        Create a new empty Directory Entry
        """
        self = cls(directory)
        self.file_number = file_number
        self.filename = ""
        self.extension = ""
        self.raw_data_link = 0
        return self

    @classmethod
    def read(
        cls, directory: "ADSSDirectory", words: t.List[int], file_number: int, position: int
    ) -> "ADSSDirectoryEntry":
        self = cls(directory)
        self.file_number = file_number
        self.filename = sixbit_to_ascii(words[position]) + sixbit_to_ascii(words[position + 1])
        self.extension = sixbit_to_ascii(words[position + 2])
        self.raw_data_link = words[position + 3]
        return self

    def to_words(self) -> t.List[int]:
        """
        Dump the directory entry to words
        """
        return [
            ascii_to_sixbit(self.filename[0:3]),
            ascii_to_sixbit(self.filename[3:6]),
            ascii_to_sixbit(self.extension),
            self.raw_data_link,
        ]

    @property
    def block_number(self) -> int:
        """
        Get the first block number of the file
        """
        return self.raw_data_link & ~0o400000

    @block_number.setter
    def block_number(self, value: int) -> None:
        """
        Set the first block number of the file
        """
        if not (0 <= value <= 0o777777):
            raise ValueError("Block number must be between 0 and 0777777")
        self.raw_data_link = (self.raw_data_link & 0o400000) | value

    @property
    def is_empty(self) -> bool:
        """
        Is the file active?
        """
        return self.raw_data_link == 0o777777 or self.raw_data_link & (1 << 17) == 0

    @is_empty.setter
    def is_empty(self, value: bool) -> None:
        """
        Set the file as empty or active
        """
        if value:
            self.raw_data_link = self.raw_data_link & ~0o400000  # Clear the active bit
        else:
            self.raw_data_link = self.raw_data_link | 0o400000  # Set the active bit

    @property
    def is_sys(self) -> bool:
        """
        Is the file a system file?
        """
        return self.directory.fs.is_system_tape and self.file_number >= SYSTEM_TAPE_DIRECTORY_ENTRIES

    @property
    def fullname(self) -> str:
        return self.basename

    @property
    def basename(self) -> str:
        return f"{self.filename};{self.extension}"

    def get_file_bitmap(self) -> "ADSSBitmap":
        """
        The File Bitmap specifies the blocks occupied by the file
        """
        return ADSSBitmap.read(self.directory.fs, self.file_number)

    def get_block_contents(self) -> t.Iterator[BlockContent]:
        """
        Get the blocks used by the file

        Normal files are stored as linked blocks.
        The last word of each data block contains the number of next block in the file.
        System programs are stored in contiguous blocks.
        """
        if self.is_sys:
            # System programs
            command_entry = self.get_command_entry()
            if command_entry:
                yield from command_entry.get_block_contents()
            else:
                block = BlockDirection(self.block_number, Direction.FORWARD)
                words = self.directory.fs.read_words_block(block)
                yield from [BlockContent(block, words)]
        else:
            direction = Direction.FORWARD
            if not self.is_empty and not self.is_sys:
                block = BlockDirection(self.block_number, direction)
                while block.block_number not in (0, 0o777777):  # 0o777777 is the end of the file
                    words = self.directory.fs.read_words_block(block)
                    yield BlockContent(block, words)
                    next_block_number = words[-1]  # Last word is the next block number
                    if direction == Direction.FORWARD and (next_block_number < block.block_number):
                        direction = Direction.BACKWARD
                    elif direction == Direction.BACKWARD and (next_block_number > block.block_number):
                        direction = Direction.FORWARD
                    block = BlockDirection(next_block_number, direction)

    def get_blocks(self) -> t.Iterator[BlockDirection]:
        """
        Get the blocks used by the file
        """
        if self.is_empty:
            pass
        elif self.is_sys:
            command_entry = self.get_command_entry()
            if command_entry:
                yield from command_entry.get_blocks()
            else:
                yield BlockDirection(self.block_number, Direction.FORWARD)
        else:
            for block_content in self.get_block_contents():
                yield block_content.block

    def get_length(self) -> int:
        """
        Get the length in blocks
        """
        if self.is_empty:
            return 0
        elif self.is_sys:
            command_entry = self.get_command_entry()
            return command_entry.length if command_entry else 1
        else:
            return self.get_file_bitmap().used()

    def get_size(self) -> int:
        """
        Get file size in bytes
        """
        return self.get_length() * self.get_block_size()

    def get_block_size(self) -> int:
        """
        Get file block size in bytes
        """
        if self.is_sys:
            return (WORDS_PER_BLOCK) * 3  # System programs are stored in contiguous blocks
        else:
            return (WORDS_PER_BLOCK - 1) * 3  # Stored as linked blocks

    def get_command_entry(self) -> t.Optional["CommandTableEntry"]:
        """
        Get the command entry associated with this directory entry
        """
        if self.is_empty or not self.is_sys:
            return None
        try:
            return self.directory.command_table.get_command_entry(self.fullname)
        except Exception:
            return None

    def delete(self) -> bool:
        """
        Delete the directory entry
        """
        # Update the directory bitmap
        bitmap = ADSSBitmap.read(self.fs)
        for block in self.get_blocks():
            bitmap.set_free(block.block_number)
        bitmap.write()
        # Update the file bitmap
        file_bitmap = ADSSBitmap.new(self.fs, self.file_number)
        file_bitmap.write()
        # Update the directory entry
        self.block_number = 0
        self.is_empty = True
        self.write()
        return True

    def write(self) -> bool:
        """
        Write the directory entry
        """
        self.directory.write()
        return True

    def allocate(self, number_of_blocks: int) -> t.List[BlockDirection]:
        """
        Allocate blocks for the file and update the directory entry
        """
        # Allocate blocks in the directory bitmap
        if number_of_blocks == 0:
            number_of_blocks = 1  # Allocate at least one block
        bitmap = ADSSBitmap.read(self.fs)
        blocks = bitmap.allocate(number_of_blocks)
        bitmap.write()
        # Update the file bitmap
        file_bitmap = ADSSBitmap.new(self.fs, self.file_number)
        for block in blocks:
            file_bitmap.set_used(block.block_number)
        file_bitmap.write()
        for i, block in enumerate(blocks):
            words = [0] * WORDS_PER_BLOCK
            next_block_number = blocks[i + 1].block_number if i + 1 < len(blocks) else 0o777777
            words[-1] = next_block_number  # Last word is the next block number
            self.fs.write_words_block(block, words)
        # Update the directory entry
        self.block_number = blocks[0].block_number
        self.is_empty = False
        self.write()
        return blocks

    def open(self, file_type: t.Optional[str] = None) -> ADSSFile:
        """
        Open a file
        """
        return ADSSFile(self, file_type)

    def read_bytes(self, file_mode: t.Optional[str] = None) -> bytes:
        """Get the content of the file"""
        file_mode = file_mode or get_file_mode(self.basename)
        result = bytearray()
        for block in self.get_block_contents():
            if self.is_sys:
                result.extend(from_18bit_words_to_bytes(block.words, IMAGE))
            else:
                result.extend(decode_block_format(block.words))
        return bytes(result)

    @property
    def fs(self) -> "ADSSFilesystem":
        return self.directory.fs

    def __str__(self) -> str:
        return f"{self.file_number:>2}  {self.filename:<6} {self.extension:<3}  {self.raw_data_link:06o} {'N' if self.is_empty else 'Y'}  {self.block_number:4}"


class ADSSDirectory:
    """
    Directory

    Word

        +-----------------------------------+
      0 | Directory Bitmap                  |  32 Words
        /                                   /
     31 |                                   |
        +-----------------------------------+
     32 | Directory Entry 1                 |   4 Words
     35 |                                   |
        +-----------------------------------+
        |                                   |
        /                                   /
        |                                   |
        +-----------------------------------+
    252 | Directory Entry 55                |   4 Words
    255 |                                   |
        +-----------------------------------+

    Pag 81
    https://bitsavers.org/pdf/dec/pdp15/DEC-15-MR2B-D_AdvMonPgmRef.pdf
    """

    fs: "ADSSFilesystem"
    entries: t.List["ADSSDirectoryEntry"]
    command_table: "CommandTable"  # Command table for system tapes

    def __init__(self, fs: "ADSSFilesystem"):
        self.fs = fs

    @classmethod
    def new(cls, fs: "ADSSFilesystem") -> "ADSSDirectory":
        """
        Create a new empty Directory
        """
        self = cls(fs)
        self.command_table = CommandTable.new(fs)
        self.entries = [ADSSDirectoryEntry.new(self, i) for i in range(DIRECTORY_ENTRIES)]
        return self

    @classmethod
    def read(cls, fs: "ADSSFilesystem") -> "ADSSDirectory":
        """
        Read Directory entries
        """
        self = cls(fs)
        # Read the directory entries
        self.entries = []
        words = self.fs.read_words_block(DIRECTORY_BLOCK)
        for i, position in enumerate(range(BITMAP_WORDS, len(words), DIRECTORY_ENTRY_WORDS)):
            entry = ADSSDirectoryEntry.read(self, words, i, position)
            self.entries.append(entry)
        # On a system tape, read the command table
        if self.fs.is_system_tape:
            try:
                sysblk = self.get_file_entry(SYSBLK_SYS)
                self.command_table = CommandTable.read(self.fs, sysblk.block_number)
            except FileNotFoundError:
                self.command_table = CommandTable.new(fs)
        return self

    def write(self) -> None:
        """
        Write the Directory to the disk
        """
        words = self.fs.read_words_block(DIRECTORY_BLOCK)[:BITMAP_WORDS]
        for entry in self.entries:
            words += entry.to_words()
        assert len(words) == 256
        self.fs.write_words_block(DIRECTORY_BLOCK, words)

    def get_file_entry(self, fullname: str) -> "ADSSDirectoryEntry":
        """
        Get a file entry by its fullname
        """
        fullname = adss_canonical_filename(fullname, wildcard=False)
        for entry in self.entries:
            if entry.fullname == fullname and not entry.is_empty:
                return entry
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), fullname)

    def get_free_entry(self) -> t.Optional[ADSSDirectoryEntry]:
        """
        Get the first free entry in the directory
        """
        for entry in self.entries:
            if entry.is_empty:
                return entry
        return None


class CommandTableEntry:
    """
    Command table entry

    Word

        +-----------------------------------+
      0 | File name                         |
      1 |                                   |
        +-----------------------------------+
      2 | Block number                      |
        +-----------------------------------+
      3 | Size                              |
        +-----------------------------------+
      4 | Load address                      |
        +-----------------------------------+
      5 | Program size                      |
        +-----------------------------------+
      6 | Start address                     |
        +-----------------------------------+

    """

    command_table: "CommandTable"
    file_number: int  # File number in the directory
    is_empty: bool
    filename: str  # Filename
    block_number: int  # First file block number
    length: int  # Length in blocks
    load_address: int
    program_size: int
    start_address: int

    def __init__(self, command_table: "CommandTable"):
        self.command_table = command_table

    @classmethod
    def read(
        cls, command_table: "CommandTable", words: t.List[int], file_number: int, position: int
    ) -> "CommandTableEntry":
        self = cls(command_table)
        self.file_number = file_number
        self.is_empty = not bool(words[position])
        self.filename = sixbit_to_ascii(words[position]) + sixbit_to_ascii(words[position + 1])
        self.block_number = words[position + 2]
        self.length = words[position + 3]
        self.load_address = words[position + 4]
        self.program_size = words[position + 5]
        self.start_address = words[position + 6]
        return self

    @property
    def fullname(self) -> str:
        return self.basename

    @property
    def basename(self) -> str:
        return f"{self.filename};SYS"

    def get_block_contents(self) -> t.Iterator[BlockContent]:
        """
        Get the blocks used by the system program
        System programs are stored in contiguous blocks.
        """
        for block in self.get_blocks():
            words = self.fs.read_words_block(block)
            yield BlockContent(block, words)

    def get_blocks(self) -> t.Iterator[BlockDirection]:
        """
        Get the blocks used by the file

        The last word of each data block contains the number of next block in the file.
        """
        yield from [
            BlockDirection(x, Direction.FORWARD) for x in range(self.block_number, self.block_number + self.length)
        ]

    def get_length(self) -> int:
        """
        Get the length in blocks
        """
        return self.length

    def get_size(self) -> int:
        """
        Get file size in bytes
        """
        return self.get_length() * self.get_block_size()

    def get_block_size(self) -> int:
        """
        Get file block size in bytes
        """
        return (WORDS_PER_BLOCK) * 3

    def delete(self) -> bool:
        """
        Delete the command entry
        """
        return False

    def write(self) -> bool:
        """
        Write the command entry
        """
        return False

    @property
    def fs(self) -> "ADSSFilesystem":
        return self.command_table.fs

    def __str__(self) -> str:
        return (
            f"{self.file_number:>2}  {self.filename:<6}  {self.block_number:<3}  {self.length:>5}  "
            f"{'N' if self.is_empty else 'Y'} {self.load_address:06o}  {self.program_size:06o}  {self.start_address:06o}"
        )


class CommandTable:
    """
    Command Table

    The Command Table is used to store additional information about system programs.

    Word

        +-----------------------------------+
      0 | Command table size                |   1 Word
        +-----------------------------------+
      2 | Command Entry 1                   |   7 Words
      6 |                                   |
        +-----------------------------------+
        |                                   |
        /                                   /
        |                                   |
        +-----------------------------------+

    """

    fs: "ADSSFilesystem"
    entries: t.List["CommandTableEntry"]

    def __init__(self, fs: "ADSSFilesystem"):
        self.fs = fs

    @classmethod
    def new(cls, fs: "ADSSFilesystem") -> "CommandTable":
        """
        Create a new empty Command Table
        """
        self = cls(fs)
        self.entries = []
        return self

    @classmethod
    def read(cls, fs: "ADSSFilesystem", block_number: t.Optional[int] = None) -> "CommandTable":
        """
        Read Command Table entries
        """
        self = cls(fs)
        self.entries = []
        if block_number is None:
            # Get the command table block number from SYSBLK;SYS
            sysblk = self.fs.get_file_entry(SYSBLK_SYS)
            block_number = sysblk.block_number
        words = self.fs.read_words_block(block_number)
        command_table_size = words[0]
        for i, position in enumerate(range(1, command_table_size, COMMAND_TABLE_ENTRY_WORDS)):
            entry = CommandTableEntry.read(self, words, i, position)
            self.entries.append(entry)
        return self

    def get_command_entry(self, fullname: str) -> "CommandTableEntry":
        """
        Get the command by name
        """
        fullname = adss_canonical_filename(fullname, wildcard=False)
        for entry in self.entries:
            if entry.fullname == fullname and not entry.is_empty:
                return entry
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), fullname)


class ADSSBitmap:
    """
    Directory/File Bitmap

    The Directory Bitmap defines block availability.
    It occupies the first 32 words of the Directory Block (block 64).
    One bit is allocated for each DECtape block.
    When set to 1, the bit indicates that the DECtape block is occupied.

    For each file in the directory, there is an additional bitmap,
    the File Bitmap, that defines the blocks occupied by that file.
    The File Bitmap occupies blocks 57 through 63.
    Each block is divided into eight File Bitmap Blocks.

    Pag 80
    https://bitsavers.org/pdf/dec/pdp15/DEC-15-MR2B-D_AdvMonPgmRef.pdf
    """

    fs: "ADSSFilesystem"
    bitmaps: t.List[int]
    bitmap_block: int
    bitmap_position: int

    def __init__(self, fs: "ADSSFilesystem", file_number: t.Optional[int] = None):
        self.fs = fs
        if file_number is None:
            self.bitmap_block = DIRECTORY_BLOCK
            self.bitmap_position = 0
        else:
            # File bitmaps are stored in blocks 57-63
            self.bitmap_block = FILE_BITMAP_BLOCK + (file_number // BITMAPS_PER_BLOCK)
            self.bitmap_position = (file_number % BITMAPS_PER_BLOCK) * BITMAP_WORDS

    @classmethod
    def new(cls, fs: "ADSSFilesystem", file_number: t.Optional[int] = None) -> "ADSSBitmap":
        """
        Create a new empty bitmap
        """
        self = cls(fs, file_number)
        self.bitmaps = [0] * BITMAP_WORDS
        return self

    @classmethod
    def read(cls, fs: "ADSSFilesystem", file_number: t.Optional[int] = None) -> "ADSSBitmap":
        """
        Read the bitmap blocks
        """
        self = cls(fs, file_number)
        words = self.fs.read_words_block(self.bitmap_block)
        self.bitmaps = words[self.bitmap_position : self.bitmap_position + BITMAP_WORDS]
        return self

    def write(self) -> None:
        """
        Write the bitmap blocks
        """
        words = self.fs.read_words_block(self.bitmap_block)
        words[self.bitmap_position : self.bitmap_position + BITMAP_WORDS] = self.bitmaps  # Update the bitmap
        self.fs.write_words_block(self.bitmap_block, words)

    @property
    def total_bits(self) -> int:
        """
        Return the bitmap length in bit
        """
        return len(self.bitmaps) * 18

    def is_free(self, bit_index: int) -> bool:
        """
        Check if a block is free
        """
        int_index = bit_index // 18
        bit_position = bit_index % 18
        bit_value = self.bitmaps[int_index]
        return (bit_value & (1 << (17 - bit_position))) == 0

    def set_used(self, bit_index: int) -> None:
        """
        Mark a block as used
        """
        int_index = bit_index // 18
        bit_position = bit_index % 18
        self.bitmaps[int_index] |= 1 << (17 - bit_position)

    def set_free(self, bit_index: int) -> None:
        """
        Mark a block as free
        """
        int_index = bit_index // 18
        bit_position = bit_index % 18
        self.bitmaps[int_index] &= ~(1 << (17 - bit_position))

    def find_contiguous_blocks(self, size: int) -> int:
        """
        Find contiguous blocks, return the first block number
        """
        current_run = 0
        start_index = -1
        for i in range(self.total_bits - 1, -1, -1):
            if self.is_free(i):
                if current_run == 0:
                    start_index = i
                current_run += 1
                if current_run == size:
                    return start_index - size + 1
            else:
                current_run = 0
        raise OSError(errno.ENOSPC, os.strerror(errno.ENOSPC))

    def allocate_one(self, from_block: int, direction: Direction) -> BlockDirection:
        """
        Allocate one block
        """
        if direction == Direction.FORWARD:
            r = range(max(from_block, 1), self.total_bits)
        else:  # Direction.BACKWARD
            r = range(min(from_block, self.total_bits - 1), 1, -1)
        for block in r:
            if self.is_free(block):
                self.set_used(block)
                return BlockDirection(block, direction)
        raise OSError(errno.ENOSPC, os.strerror(errno.ENOSPC))

    def allocate(self, size: int) -> t.List[BlockDirection]:
        """
        Allocate blocks in a staggered manner.

        The first block is always recorded in the forward direction.
        Free blocks are chosen which are at least 5 beyond the last one.
        """
        blocks: t.List[BlockDirection] = []
        direction = Direction.FORWARD
        found = False
        while len(blocks) < size:
            try:
                if blocks:
                    pos = blocks[-1].block_number + (5 if direction == Direction.FORWARD else -5)
                else:
                    pos = 1
                block = self.allocate_one(pos, direction)
                blocks.append(block)
                found = True
            except OSError:
                if not found:  # If no block was found
                    raise
                direction = direction.reverse()
                found = False
        assert len(blocks) == size
        return blocks

    def used(self) -> int:
        """
        Count the number of used blocks
        """
        used = 0
        for block in self.bitmaps:
            used += block.bit_count()
        return used

    def get_allocated_blocks(self) -> t.List[int]:
        """
        Get the list of allocated blocks
        """
        allocated_blocks = []
        for i in range(self.total_bits):
            if not self.is_free(i):
                allocated_blocks.append(i)
        return allocated_blocks

    def free(self) -> int:
        """
        Count the number of free blocks
        """
        return self.total_bits - self.used()

    def __str__(self) -> str:
        free = self.free()
        used = self.used()
        return f"Free blocks: {free:<6} Used blocks: {used:<6}"

    def __eq__(self, other: object) -> bool:
        return isinstance(other, ADSSBitmap) and self.bitmaps == other.bitmaps  # type: ignore


class ADSSFilesystem(AbstractFilesystem):
    """
    Advanced Monitor Software System (ADSS) filesystem for PDP-9/PDP-15
    This filesystem is also used by DOS-15 on DECtapes.

    Block 64 is the Directory Block, which contains the Directory Bitmap and Directory Entries.
    The first 32 words of the Directory Block are used for the Directory Bitmap.
    Each Directory Entry occupies 4 words.
    The Directory contains 56 entries.
    If the tape is a system tape, the first 32 entries are used for user files,
    and the last 24 entries are reserved for system programs.
    User files are stored as linked blocks, while system files are stored in contiguous blocks.
    Each file as an additional File Bitmap, which occupies blocks 57-63.
    The File Bitmap defines the blocks occupied by the file.
    For the system programs, there is a Command Table that contains additional information, such as the size.
    """

    fs_name = "adss"
    fs_description = "PDP-15 Advanced Monitor Software System"
    fs_platforms = ["pdp-9", "pdp-15"]
    is_system_tape: bool = False  # System tape
    dev: BlockDevice18Bit

    def __init__(self, file_or_device: t.Union["AbstractFile", "AbstractDevice"]):
        if isinstance(file_or_device, AbstractFile):
            self.dev = BlockDevice18Bit(file_or_device, words_per_block=WORDS_PER_BLOCK)
        elif isinstance(file_or_device, BlockDevice18Bit):
            self.dev = file_or_device
        else:
            raise OSError(errno.EIO, f"Invalid device type for {self.fs_description} filesystem")

    @classmethod
    def mount(
        cls,
        file_or_dev: t.Union["AbstractFile", "AbstractDevice"],
        strict: t.Union[bool, str] = True,
        **kwargs: t.Union[bool, str],
    ) -> "ADSSFilesystem":
        """
        Mount the filesystem from a file or device
        """
        self = cls(file_or_dev)
        words = self.dev.read_words_block(DIRECTORY_BLOCK)
        # Check if the tape is a system tape
        self.is_system_tape = words[DIRECTORY_SYSTEM_WORD] == 0o777777
        if strict:
            # Check the tape size
            blocks = self.get_size() // 3 // 512
            if abs(blocks - DECTAPE_BLOCKS) >= 4:
                raise OSError(errno.EINVAL, "Invalid ADSS tape size")
            # Check the bitmap
            bitmap = ADSSBitmap.read(self)
            # Mark the File Bitmap Blocks as used
            for block_number in [DIRECTORY_BLOCK, FILE_BITMAP_BLOCK]:
                if bitmap.is_free(block_number):
                    raise OSError(errno.EINVAL, "Invalid ADSS filesystem")
        return self

    def read_words_block(
        self,
        block: t.Union[int, BlockDirection],
    ) -> t.List[int]:
        """
        Read a 256 bytes block as 18bit words

        It is possible to read data from a DECtape in backward direction.
        However, a re-ordering of both the entire block and individual words is required.

        Pag 75
        https://bitsavers.org/pdf/dec/pdp15/DEC-15-H2DC-D_usersVol2.pdf
        """
        return self.dev.read_words_block(block)

    def write_words_block(
        self,
        block: t.Union[int, BlockDirection],
        words: t.List[int],
    ) -> None:
        """
        Write 256 18bit words as a block
        """
        self.dev.write_words_block(block, words)

    def read_dir_entries(self) -> t.Iterator["ADSSDirectoryEntry"]:
        """
        Read directory entries
        """
        yield from ADSSDirectory.read(self).entries

    @property
    def entries_list(self) -> t.Iterator["ADSSDirectoryEntry"]:
        for entry in self.read_dir_entries():
            if not entry.is_empty:
                yield entry

    def filter_entries_list(
        self,
        pattern: t.Optional[str],
        include_all: bool = False,
        expand: bool = True,
        wildcard: bool = True,
    ) -> t.Iterator["ADSSDirectoryEntry"]:
        pattern = adss_canonical_filename(pattern, wildcard=True) if pattern else ""
        for entry in self.read_dir_entries():
            if not entry.is_empty and filename_match(entry.basename, pattern, wildcard):
                yield entry

    def get_file_entry(self, fullname: str) -> "ADSSDirectoryEntry":
        """
        Get the directory entry for a file
        """
        for entry in self.filter_entries_list(fullname, wildcard=False):
            return entry
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), fullname)

    def create_file(
        self,
        fullname: str,
        number_of_blocks: int,  # length in blocks
        creation_date: t.Optional[date] = None,  # optional creation date
        file_type: t.Optional[str] = None,
    ) -> t.Optional["ADSSDirectoryEntry"]:
        """
        Create a new file with a given length in number of blocks
        """
        fullname = adss_canonical_filename(fullname)
        # If the file already exists, delete it
        try:
            entry = self.get_file_entry(fullname)
            entry.delete()
        except FileNotFoundError:
            # Seach for a free entry in the directory
            directory = ADSSDirectory.read(self)
            entry = directory.get_free_entry()  # type: ignore
            if entry is None:
                raise OSError(errno.ENOSPC, "Directory full")
            filename, extension = adss_split_ext(fullname)
            entry.filename = filename
            entry.extension = extension
        # Allocate space and write the entry
        entry.allocate(number_of_blocks)
        return entry

    def write_bytes(
        self,
        fullname: str,
        content: bytes,
        creation_date: t.Optional[date] = None,
        file_type: t.Optional[str] = None,
        file_mode: t.Optional[str] = None,
    ) -> None:
        """
        Write content to a file
        """
        fullname = adss_canonical_filename(fullname)
        file_mode = file_mode or get_file_mode(fullname)
        blocks_content = list(encode_block_format(content, file_mode, words_per_block=WORDS_PER_BLOCK - 1))
        number_of_blocks = len(blocks_content)
        # Create the file entry
        entry = self.create_file(fullname, number_of_blocks, creation_date, file_type)
        # Write the file content
        if entry is not None:
            for block, block_content in zip(entry.get_block_contents(), blocks_content):
                assert len(block_content) == WORDS_PER_BLOCK - 1
                block.words[: WORDS_PER_BLOCK - 1] = block_content
                self.write_words_block(block.block, block.words)

    def dir(self, volume_id: str, pattern: t.Optional[str], options: t.Dict[str, bool]) -> None:
        entries = self.filter_entries_list(pattern, wildcard=True)
        if not entries:
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), pattern)
        if not options.get("brief"):
            sys.stdout.write("DIRECTORY LISTING\n")
        for x in entries:
            if not x.is_empty:
                if options.get("brief"):
                    sys.stdout.write(f"{x.filename:<6};{x.extension:<3}\n")
                else:
                    sys.stdout.write(f"{x.filename:<6} {x.extension:<3}  {x.block_number:>4o}\n")
        if not options.get("brief"):
            bitmap = ADSSBitmap.read(self)
            sys.stdout.write(f"{bitmap.free():<4o} FREE BLOCKS\n")

    def examine(self, arg: t.Optional[str], options: t.Dict[str, t.Union[bool, str]]) -> None:
        if options.get("bitmap"):
            # Display the bitmap
            file_number = int(arg) if arg and arg.isdigit() else None
            bitmap = ADSSBitmap.read(self, file_number)
            for i in range(0, bitmap.total_bits):
                sys.stdout.write(f"{i:>4d} {'[ ]' if bitmap.is_free(i) else '[X]'}  ")
                if i % 16 == 15:
                    sys.stdout.write("\n")
            sys.stdout.write(f"\nUsed blocks: {bitmap.used()}\n")
        elif arg:
            # Display the file entry
            entries = self.filter_entries_list(arg, wildcard=True)
            for entry in entries:
                sys.stdout.write(f"File number:              {entry.file_number}\n")
                sys.stdout.write(f"Filename:                 {entry.filename}\n")
                sys.stdout.write(f"Extension:                {entry.extension}\n")
                sys.stdout.write(f"Raw data link:            {entry.raw_data_link:06o}\n")
                sys.stdout.write(f"Empty:                    {'Y' if entry.is_empty else 'N'}\n")
                sys.stdout.write(f"System file:              {'Y' if entry.is_sys else 'N'}\n")
                sys.stdout.write(f"First block number:       {entry.block_number}\n")
                sys.stdout.write(f"Size:                     {entry.get_length()}\n")
                command_entry = entry.get_command_entry()
                if command_entry:
                    sys.stdout.write(f"Command entry number:     {command_entry.file_number}\n")
                    sys.stdout.write(f"Command block number:     {command_entry.block_number}\n")
                    sys.stdout.write(f"Load address:             {command_entry.load_address:06o}\n")
                    sys.stdout.write(f"Program size:             {command_entry.program_size:06o}\n")
                    sys.stdout.write(f"Start address:            {command_entry.start_address:06o}\n")
                bitmap = ADSSBitmap.read(self, entry.file_number)
                blocks = str(list(entry.get_blocks()))
                sys.stdout.write(f"Blocks:                   {blocks}\n")
                sys.stdout.write("\n")
                if entry.basename == SYSBLK_SYS:
                    # Display the command table
                    sys.stdout.write("Command Table:\n\n")
                    sys.stdout.write("Num  Filename  Block  Length  Load Addr  Prog Size  Start Addr\n")
                    sys.stdout.write("---  --------  -----  ------  ---------  ---------  ----------\n")
                    for command_entry in entry.directory.command_table.entries:
                        sys.stdout.write(
                            f"{command_entry.file_number:>2}   {command_entry.filename:<6}   "
                            f"{command_entry.block_number:>6}  {command_entry.length:>6}     "
                            f"{command_entry.load_address:06o}     {command_entry.program_size:06o}      "
                            f"{command_entry.start_address:06o}\n"
                        )
                        # f"{'N' if command_entry.is_empty else 'Y'} "
        else:
            entries = self.read_dir_entries()
            directory = ADSSDirectory.read(self)
            bitmap = ADSSBitmap.read(self)
            user_files = len([x for x in directory.entries if not x.is_empty and not x.is_sys])
            user_blocks = sum([x.get_length() for x in directory.entries if not x.is_empty and not x.is_sys])
            system_blocks = bitmap.used() - user_blocks
            sys.stdout.write(f"Free blocks:           {bitmap.free():>4}\n")
            sys.stdout.write(f"User files:            {user_files:>4}\n")
            sys.stdout.write(f"System blocks:         {system_blocks:>4}\n")
            sys.stdout.write(f"Is system tape:           {'Y' if self.is_system_tape else 'N'}\n")
            sys.stdout.write("\n")
            # Display the directory entries
            sys.stdout.write("Num   Filename  Data link Block  Size  Type\n")
            sys.stdout.write("---   --------  --------- -----  ----  ----\n")
            full = bool(options.get("full", False))
            for entry in directory.entries:
                if full or not entry.is_empty:
                    if entry.is_empty:
                        ftype = '  -'
                    elif entry.is_sys:
                        ftype = 'System'
                    elif get_file_mode(entry.basename) == IMAGE:
                        ftype = 'Binary'
                    else:
                        ftype = 'ASCII'
                    block_number = f"{entry.block_number:4d}" if not entry.is_empty else "-"
                    length = f"{entry.get_length():4d}" if not entry.is_empty else "-"
                    sys.stdout.write(
                        f"{entry.file_number:>2}  {entry.filename:>6};{entry.extension:<3}  "
                        f"{entry.raw_data_link:06o} {block_number:>8}  {length:>4}  {ftype}\n"
                    )

    def dump(self, fullname: t.Optional[str], start: t.Optional[int] = None, end: t.Optional[int] = None) -> None:
        """
        Dump the content of a file or a range of blocks
        """
        if fullname:
            entry = self.get_file_entry(fullname)
            for i, block_content in enumerate(entry.get_block_contents()):
                if (start is None or start <= i) and (end is None or i <= end):
                    sys.stdout.write(f"\nBLOCK NUMBER   {str(block_content.block):>8} ({i:08})\n")
                    oct_dump(block_content.words)
        else:
            if start is None:
                start = 0
                if end is None:  # full disk
                    end = (self.get_size() // WORDS_PER_BLOCK // 4) - 1
            elif end is None:  # one single block
                end = start
            for block_number in range(start, end + 1):
                words = self.read_words_block(block_number)
                sys.stdout.write(f"\nBLOCK NUMBER   {block_number:>8}\n")
                oct_dump(words)

    @classmethod
    def initialize(
        cls, file_or_dev: t.Union["AbstractFile", "AbstractDevice"], **kwargs: t.Union[bool, str]
    ) -> "ADSSFilesystem":
        """
        Create an empty filesystem
        """
        self = cls(file_or_dev)
        self.is_system_tape = False
        bitmap = ADSSBitmap.new(self)
        # Mark the File Bitmap Blocks as used
        for block_number in range(FILE_BITMAP_BLOCK, FILE_BITMAP_BLOCK + BITMAPS_PER_BLOCK):
            bitmap.set_used(block_number)
        # Mark the Directory Block as used
        bitmap.set_used(DIRECTORY_BLOCK)
        bitmap.write()
        # Create an empty directory
        ADSSDirectory.new(self).write()
        return self

    def get_size(self) -> int:
        """
        Get filesystem size in bytes
        """
        return self.dev.get_size()



# Copyright (C) 2014 Andrea Bonomi <andrea.bonomi@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import errno
import math
import os
import struct
import sys
import typing as t
from datetime import date

from ..abstract import AbstractDirectoryEntry, AbstractFile, AbstractFilesystem
from ..commons import BLOCK_SIZE, READ_FILE_FULL, filename_match
from ..device.abstract import AbstractDevice
from ..device.tape import Tape
from .rt11fs import rt11_canonical_filename

__all__ = [
    "CAPS11File",
    "CAPS11DirectoryEntry",
    "CAPS11Filesystem",
]

HEADER_RECORD = ">6s3sBHBB6sB11s"
HEADER_RECORD_SIZE = 32
RECORD_SIZE = 128
SENTINEL_FILE = b"\0" * HEADER_RECORD_SIZE

# CAPS-11 uses file type codes 0o1 (ASCII), 0o2 (BIN), and 0o14 (BAD)
# CAPS-11 Users Guide, Pag 290
# http://bitsavers.org/pdf/dec/pdp11/caps-11/DEC-11-OTUGA-A-D_CAPS-11_Users_Guide_Oct73.pdf

FILE_TYPE_ASCII = 0o1  # ASCII (seven bits per character)
FILE_TYPE_BIN = 0o2
FILE_TYPE_CORE1 = 0o3  # One 36-bit word in 5 bytes
FILE_TYPE_CORE2 = 0o4  # One 12-bit word in 2 bytes
FILE_TYPE_CORE3 = 0o5  # One 18-bit word in 3 bytes
FILE_TYPE_CORE4 = 0o6  # One 36-bit word in 6 bytes
FILE_TYPE_CORE5 = 0o7  # One 16-bit word in 2 bytes
FILE_TYPE_CORE6 = 0o10  # 2 x 12-bit words in 3 bytes
FILE_TYPE_CORE7 = 0o11  # 2 x 36-bit words in 9 bytes
FILE_TYPE_CORE8 = 0o12  # 4 x 18-bit words in 9 bytes
FILE_TYPE_BOOT = 0o13  # Bootstrap
FILE_TYPE_BAD = 0o14  # Bad file

STANDARD_FILE_TYPES = {
    FILE_TYPE_ASCII: "ASCII",
    FILE_TYPE_BIN: "BIN",
    FILE_TYPE_CORE1: "CORE1",
    FILE_TYPE_CORE2: "CORE2",
    FILE_TYPE_CORE3: "CORE3",
    FILE_TYPE_CORE4: "CORE4",
    FILE_TYPE_CORE5: "CORE5",
    FILE_TYPE_CORE6: "CORE6",
    FILE_TYPE_CORE7: "CORE7",
    FILE_TYPE_CORE8: "CORE8",
    FILE_TYPE_BOOT: "BOOT",
    FILE_TYPE_BAD: "BAD",
}


def caps11_to_date(val: bytes) -> t.Optional[date]:
    """
    Translate CAPS-11 date (stored in ASCII as ddmmyy) to Python date
    """
    try:
        date_str = val.decode("ascii", errors="ignore")
        day = int(date_str[0:2])
        month = int(date_str[2:4])
        year = int(date_str[4:6])
        return date(year + (1900 if year >= 60 else 2000), month, day)
    except:
        return None


def date_to_caps11(d: t.Optional[date]) -> bytes:
    """
    Translate Python date to CAPS-11 date
    """
    if d is None:
        return b"     "
    day = f"{d.day:02}"
    month = f"{d.month:02}"
    year = f"{d.year:04}"[2:4]
    date_str = f"{day}{month}{year}"
    return date_str.encode("ascii")


class CAPS11File(AbstractFile):
    entry: "CAPS11DirectoryEntry"
    closed: bool
    size: int  # size in bytes
    content: bytes  # file content

    def __init__(self, entry: "CAPS11DirectoryEntry"):
        self.entry = entry
        self.closed = False
        self.size = entry.size
        entry.fs.dev.tape_seek(entry.tape_pos)
        self.content = entry.fs.dev.tape_read_file()[HEADER_RECORD_SIZE + entry.continued :]

    def read_block(
        self,
        block_number: int,
        number_of_blocks: int = 1,
    ) -> bytes:
        """
        Read block(s) of data from the file
        """
        if number_of_blocks == READ_FILE_FULL:
            number_of_blocks = self.entry.get_length()
        if (
            self.closed
            or block_number < 0
            or number_of_blocks < 0
            or block_number + number_of_blocks > self.entry.get_length()
        ):
            raise OSError(errno.EIO, os.strerror(errno.EIO))
        return self.content[block_number * BLOCK_SIZE : (block_number + number_of_blocks) * BLOCK_SIZE]

    def write_block(
        self,
        buffer: bytes,
        block_number: int,
        number_of_blocks: int = 1,
    ) -> None:
        """
        Write block(s) of data to the file
        """
        raise OSError(errno.EROFS, os.strerror(errno.EROFS))

    def get_size(self) -> int:
        """
        Get file size in bytes
        """
        return self.size

    def get_block_size(self) -> int:
        """
        Get file block size in bytes
        """
        return BLOCK_SIZE

    def close(self) -> None:
        """
        Close the file
        """
        self.closed = True

    def __str__(self) -> str:
        return self.entry.fullname


class CAPS11DirectoryEntry(AbstractDirectoryEntry):
    """
    CAPS-11 File Header

      9 bytes   1 byte  2 bytes        1 byte 1 byte  6 bytes  12 bytes
    +----------+------+---------------+------+-------+--------+--------+
    | Filename | Type | Record length | Seq. | Cont. |  Data  |  Spare |
    +----------+------+---------------+------+-------+--------+--------+

    CAPS-11 Users Guide, Pag 289
    http://bitsavers.org/pdf/dec/pdp11/caps-11/DEC-11-OTUGA-A-D_CAPS-11_Users_Guide_Oct73.pdf

    CAPS-8 File Header

      9 bytes   1 byte  2 bytes        1 byte  1 byte  6 bytes   1 byte  11 bytes
    +----------+------+---------------+------+-------+--------+---------+---------+
    | Filename | Type | Record length | Seq. | Cont. |  Data  | Version |  Spare  |
    +----------+------+---------------+------+-------+--------+---------+---------+

    CAPS-8 Users Guide, Pag 205
    https://bitsavers.org/pdf/dec/pdp8/caps8/DEC-8E-OCASA-B-D_CAPS8_UG.pdf

    """

    fs: "CAPS11Filesystem"
    filename: str = ""  #             6 chars - name
    extension: str = ""  #            3 chars - extension
    record_type: int = 0  #           1 byte  - record type
    record_length: int = 0  #         2 bytes - file record length (fixed at 128 bytes)
    sequence: int = 0  #              1 byte  - file sequence number for multi volume files (0)
    continued: int = 0  #             1 byte  - header auxiliary header record (0)
    raw_creation_date: bytes = b""  # 6 char  - file creation date as DDMMYY
    version: int = 0  #               1 byte  - version number (incremented by the editor) - CAPS-8 only
    unused: bytes = b""  #           11 char  - spare

    file_number: int = 0
    size: int = 0  # size in bytes
    tape_pos: int = 0  # tape position (before file header)

    def __init__(self, fs: "CAPS11Filesystem"):
        self.fs = fs

    @classmethod
    def new(
        cls,
        fs: "CAPS11Filesystem",
        file_number: int,
        tape_pos: int,
        filename: str,
        extension: str,
        creation_date: t.Optional[date] = None,  # optional creation date
        record_type: int = FILE_TYPE_BIN,
    ) -> "CAPS11DirectoryEntry":
        self = CAPS11DirectoryEntry(fs)
        self.file_number = file_number
        self.tape_pos = tape_pos
        self.filename = filename
        self.extension = extension
        self.record_type = record_type
        self.record_length = RECORD_SIZE
        self.raw_creation_date = date_to_caps11(creation_date)
        self.sequence = 0
        self.continued = 0
        self.version = 0
        self.unused = b"\0" * 11
        return self

    @classmethod
    def read(
        cls,
        fs: "CAPS11Filesystem",
        buffer: bytes,
        file_number: int,
        tape_pos: int,
        size: int,
    ) -> "CAPS11DirectoryEntry":
        self = CAPS11DirectoryEntry(fs)
        self.file_number = file_number
        self.tape_pos = tape_pos
        (
            filename,
            extension,
            self.record_type,
            self.record_length,
            self.sequence,
            self.continued,
            self.raw_creation_date,
            self.version,
            self.unused,
        ) = struct.unpack_from(HEADER_RECORD, buffer, 0)
        # print(" ".join([f"{b:02x}" for b in buffer]))
        self.filename = filename.decode("ascii", errors="ignore").rstrip(" ")
        self.extension = extension.decode("ascii", errors="ignore").rstrip(" ")
        if not self.filename or self.filename.startswith("\0"):  # Sentinel file
            self.filename = ""
            self.extension = ""
        self.size = size - self.continued
        return self

    def write(self, skip_file: bool = True) -> bool:
        """
        Write the directory entry
        """
        buffer = bytearray(HEADER_RECORD_SIZE)
        if not self.filename:  # Sentinel file
            filename = b"\0"
            extension = b"\0"
        else:
            filename = self.filename.ljust(6).encode("ascii", errors="ignore")
            extension = self.extension.ljust(3).encode("ascii", errors="ignore")
        # Pack the data into the buffer
        struct.pack_into(
            HEADER_RECORD,
            buffer,
            0,
            filename,
            extension,
            self.record_type,
            self.record_length,
            self.sequence,
            self.continued,
            self.raw_creation_date,
            self.version,
            self.unused,
        )
        self.fs.dev.tape_seek(self.tape_pos)
        self.fs.dev.tape_write_forward(buffer)
        if skip_file:
            self.fs.dev.tape_skip_file()
        return True

    def get_length(self) -> int:
        """
        Get the length in blocks
        """
        return int(math.ceil(self.size / BLOCK_SIZE))

    def get_size(self) -> int:
        """
        Get file size in bytes
        """
        return self.size

    def get_block_size(self) -> int:
        """
        Get file block size in bytes
        """
        return BLOCK_SIZE

    @property
    def is_empty(self) -> bool:
        return self.record_type == FILE_TYPE_BAD or self.is_sentinel_file

    @property
    def is_sentinel_file(self) -> bool:
        return not self.filename

    @property
    def fullname(self) -> str:
        return f"{self.filename}.{self.extension}"

    @property
    def basename(self) -> str:
        return f"{self.filename}.{self.extension}"

    @property
    def creation_date(self) -> t.Optional[date]:
        return caps11_to_date(self.raw_creation_date)

    @property
    def file_type(self) -> t.Optional[str]:
        return STANDARD_FILE_TYPES.get(self.record_type)

    def delete(self) -> bool:
        """
        Delete the file
        """
        self.filename = "*EMPTY"
        self.extension = ""
        self.record_type = FILE_TYPE_BAD
        self.record_length = 0
        self.sequence = 0
        self.continued = 0
        self.raw_creation_date = b""
        self.write()
        return True

    def open(self, file_mode: t.Optional[str] = None) -> CAPS11File:
        """
        Open a file
        """
        return CAPS11File(self)

    def __str__(self) -> str:
        file_type = self.file_type or f"{self.record_type:>4o}"
        return (
            f"{self.file_number:<4} "
            f"{self.filename:>6}."
            f"{self.extension:<3}  "
            f"{file_type:>6}  "
            f"{self.record_length:>6} "
            f"{self.sequence:>4} "
            f"{self.continued:>4}  "
            f"{self.creation_date or '          '} "
            f"{self.size:>8} "
            f"{(self.version if self.fs.caps8 else ''):>3}"
        )

    def __repr__(self) -> str:
        return str(self)


class CAPS11Filesystem(AbstractFilesystem):
    """
    CAPS-11 Filesystem

        +-------------------------------------+
     1  |           File header               |  32 bytes
        +-------------------------------------+
     2  |              Data                   | 128 bytes
        +-------------------------------------+
     n  |               ...                   |
        +-------------------------------------+
    n-1 |              Data                   |
        +-------------------------------------+
     n  |               EOF                   |
        +-------------------------------------+

    CAPS-11 Users Guide, Pag 287
    http://bitsavers.org/pdf/dec/pdp11/caps-11/DEC-11-OTUGA-A-D_CAPS-11_Users_Guide_Oct73.pdf
    """

    fs_name = "caps11"
    fs_description = "PDP-11 CAPS-11 / PDP-8 CAPS-8"
    fs_platforms = ["pdp11", "pdp8"]
    caps8 = False  # Is CAPS-8 ?
    dev: Tape

    def __init__(self, file_or_device: t.Union["AbstractFile", "AbstractDevice"]):
        if isinstance(file_or_device, AbstractFile):
            self.dev = Tape(file_or_device)
        elif isinstance(file_or_device, Tape):
            self.dev = file_or_device
        else:
            raise OSError(errno.EIO, f"Invalid device type for {self.fs_description} filesystem")

    @classmethod
    def mount(
        cls,
        file_or_dev: t.Union["AbstractFile", "AbstractDevice"],
        strict: t.Union[bool, str] = True,
        **kwargs: t.Union[bool, str],
    ) -> "AbstractFilesystem":
        self = cls(file_or_dev)
        if strict:
            for entry in self.read_file_headers(include_eot=False):
                if not entry.is_sentinel_file and entry.version != 0:
                    self.caps8 = True
                if entry.record_length not in [0, RECORD_SIZE] or entry.size < 0:
                    raise OSError(errno.EIO, f"Invalid record length ({entry.record_length}) {entry}")
        return self

    def read_file_headers(self, include_eot: bool = False) -> t.Iterator["CAPS11DirectoryEntry"]:
        """Read file headers"""
        self.dev.tape_rewind()
        try:
            file_number = 0
            while True:
                tape_pos = self.dev.tape_pos
                header, size = self.dev.tape_read_header()
                if header:
                    file_number += 1
                    entry = CAPS11DirectoryEntry.read(self, header, file_number, tape_pos, size)
                    if include_eot or not entry.is_sentinel_file:
                        yield entry
        except EOFError:
            pass

    def filter_entries_list(
        self,
        pattern: t.Optional[str],
        include_all: bool = False,
        expand: bool = True,
        wildcard: bool = True,
    ) -> t.Iterator["CAPS11DirectoryEntry"]:
        if pattern:
            pattern = rt11_canonical_filename(pattern, wildcard=wildcard)
        for entry in self.read_file_headers():
            if filename_match(entry.basename, pattern, wildcard) and (include_all or not entry.is_empty):
                yield entry

    @property
    def entries_list(self) -> t.Iterator["CAPS11DirectoryEntry"]:
        for entry in self.read_file_headers():
            if not entry.is_empty:
                yield entry

    def get_file_entry(self, fullname: str) -> CAPS11DirectoryEntry:
        fullname = rt11_canonical_filename(fullname)
        if not fullname:
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), fullname)
        try:
            return next(self.filter_entries_list(fullname, wildcard=False))
        except StopIteration:
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), fullname)

    def write_bytes(
        self,
        fullname: str,
        content: bytes,
        creation_date: t.Optional[date] = None,
        file_type: t.Optional[str] = None,
        file_mode: t.Optional[str] = None,
    ) -> None:
        """
        Write content to a file
        """
        number_of_blocks = int(math.ceil(len(content) * 1.0 / RECORD_SIZE))
        self.create_file(fullname, number_of_blocks, creation_date, content=content)

    def create_file(
        self,
        fullname: str,
        number_of_blocks: int,  # length in blocks
        creation_date: t.Optional[date] = None,  # optional creation date
        file_type: t.Optional[str] = None,
        content: t.Optional[bytes] = None,
    ) -> t.Optional[CAPS11DirectoryEntry]:
        """
        Create a new file with a given length in number of blocks
        """
        # Delete the existing file
        fullname = rt11_canonical_filename(fullname, wildcard=False)
        try:
            self.get_file_entry(fullname).delete()
        except FileNotFoundError:
            pass
        # Find the position for the new file
        tape_pos = 0
        for entry in reversed(list(self.read_file_headers(include_eot=True))):
            if (not entry.is_sentinel_file) and entry.record_type != FILE_TYPE_BAD:
                break
            tape_pos = entry.tape_pos
        self.dev.tape_seek(tape_pos)
        self.dev.tape_truncate(tape_pos)
        # Create the new directory entry
        filename, extension = fullname.split(".", 1)
        entry = CAPS11DirectoryEntry.new(self, 0, tape_pos, filename, extension, creation_date)
        entry.write(skip_file=False)
        # Write the file
        empty_record = b"\0" * RECORD_SIZE
        for i in range(0, number_of_blocks):
            if content is not None:
                record = content[i * RECORD_SIZE : (i + 1) * RECORD_SIZE]
                if len(record) < RECORD_SIZE:
                    record += b"\0" * (RECORD_SIZE - len(record))
                self.dev.tape_write_forward(record)
            else:
                self.dev.tape_write_forward(empty_record)
        # Write tape mark
        self.dev.tape_write_mark()
        # Write the sentinel file
        self.write_sentinel_file()
        return entry

    def dir(self, volume_id: str, pattern: t.Optional[str], options: t.Dict[str, bool]) -> None:
        if not options.get("brief"):
            if self.caps8:
                dt = date.today().strftime('%m/%d/%y').upper()
                sys.stdout.write(f"{dt}\n")
            else:
                dt = date.today().strftime('%d-%B-%y').upper()
                sys.stdout.write(f" {dt}\n\n")

        for x in self.filter_entries_list(pattern, include_all=True):
            if options.get("brief"):
                # Omit creation date and version number
                if x.is_empty:
                    continue
                elif self.caps8:
                    sys.stdout.write(f"{x.filename:<6s}.{(x.extension or 'BIN'):<3s}\n")
                else:
                    sys.stdout.write(f"{x.filename:<6s} {x.extension:<3s}\n")
            elif self.caps8:
                version = f"V{x.version}" if x.version else ""
                creation_date = x.creation_date and x.creation_date.strftime("%m/%d/%y").upper() or ""
                sys.stdout.write(f"{x.filename:<6s}.{(x.extension or 'BIN'):<3s} {creation_date:<8s} {version}\n")
            else:
                creation_date = x.creation_date and x.creation_date.strftime("%d-%b-%y").upper() or "--"
                sys.stdout.write(f"{x.filename:<6s} {x.extension:<3s} {creation_date:<9s}\n")

    def examine(self, arg: t.Optional[str], options: t.Dict[str, t.Union[bool, str]]) -> None:
        if arg:
            self.dump(arg)
        else:
            if self.caps8:
                sys.stdout.write("Num    Filename    Type     Rec  Seq Cont        Date     Size Ver\n")
                sys.stdout.write("---    --------    ----     ---  --- ----        ----     ---- ---\n")
            else:
                sys.stdout.write("Num    Filename    Type     Rec  Seq Cont        Date     Size\n")
                sys.stdout.write("---    --------    ----     ---  --- ----        ----     ----\n")
            for entry in self.read_file_headers(include_eot=True):
                sys.stdout.write(f"{entry}\n")

    def get_size(self) -> int:
        """
        Get filesystem size in bytes
        """
        return self.dev.get_size()

    @classmethod
    def initialize(
        cls, file_or_dev: t.Union["AbstractFile", "AbstractDevice"], **kwargs: t.Union[bool, str]
    ) -> "CAPS11Filesystem":
        """
        Initialize the filesystem
        """
        self = cls(file_or_dev)
        self.dev.tape_rewind()
        self.dev.tape_write_mark()
        self.write_sentinel_file()
        return self

    def write_sentinel_file(self) -> None:
        """
        Write the sentinel file at the current tape position
        The sentinel file is the last file on a tape
        """
        self.dev.tape_write_forward(SENTINEL_FILE)
        self.dev.tape_truncate()

    def close(self) -> None:
        self.dev.close()

    def get_types(self) -> t.List[str]:
        """
        Get the list of the supported file types
        """
        return list(STANDARD_FILE_TYPES.values())



from datetime import date

import pytest

from xferx.pdp11.rstsfs import (  # UserFileDirectoryBlock,; date_to_rsts,; rsts_to_date,
    ANY_GROUP,
    ANY_USER,
    PPN,
    RSTSFilesystem,
)
from xferx.shell import Shell

DSK = "tests/dsk/rsts.dsk"


def test_rsts():
    shell = Shell(verbose=True)
    shell.onecmd(f"mount t: /rsts {DSK}", batch=True)
    fs = shell.volumes.get('T')
    assert isinstance(fs, RSTSFilesystem)

    shell.onecmd("dir t:", batch=True)
    shell.onecmd("dir /uic t:", batch=True)
    shell.onecmd("type t:[100,100]1.txt", batch=True)

    x = fs.read_bytes("[5,10]50.txt")
    x = x.rstrip(b"\0")
    # assert len(x) == 2200
    for i in range(0, 50):
        assert f"{i:5d} ABCDEFGHIJKLMNOPQRSTUVWXYZ01234567890".encode("ascii") in x

    l = list(fs.filter_entries_list("*.TXT[100,100]"))
    assert len(l) == 4

    x1 = fs.read_bytes("[100,100]1000.txt")
    x1 = x1.rstrip(b"\0")
    # assert len(x1) == 44000
    for i in range(0, 1000):
        assert f"{i:5d} ABCDEFGHIJKLMNOPQRSTUVWXYZ01234567890".encode("ascii") in x1

    l = list(fs.filter_entries_list("*.TXT[200,200]"))
    # assert len(l) == 3

    x2 = fs.read_bytes("[100,100]1000.txt")
    x2 = x2.rstrip(b"\0")
    # assert len(x2) == 44000
    for i in range(0, 1000):
        assert f"{i:5d} ABCDEFGHIJKLMNOPQRSTUVWXYZ01234567890".encode("ascii") in x2
    assert x1 == x2

    l = list(fs.filter_entries_list("*.TXT[1,2]"))
    assert len(l) == 0

    x3 = fs.read_bytes("[5,10]500.txt")
    x3 = x3.rstrip(b"\0")
    # assert len(x3) == 22000
    for i in range(0, 500):
        assert f"{i:5d} ABCDEFGHIJKLMNOPQRSTUVWXYZ01234567890".encode("ascii") in x3

    x4 = fs.read_bytes("[5,10]5000.txt")
    x4 = x4.rstrip(b"\0")
    for i in range(0, 5000):
        assert f"{i:5d} ABCDEFGHIJKLMNOPQRSTUVWXYZ01234567890".encode("ascii") in x4


def test_ppn_from_str_normal_case():
    ppn = PPN.from_str("[123,45]")
    assert ppn.group == 123
    assert ppn.user == 45


def test_ppn_from_str_any_group():
    ppn = PPN.from_str("[*,45]")
    assert ppn.group == ANY_GROUP
    assert ppn.user == 45


def test_ppn_from_str_any_user():
    ppn = PPN.from_str("[123,*]")
    assert ppn.group == 123
    assert ppn.user == ANY_USER


def test_ppn_from_str_any_group_and_user():
    ppn = PPN.from_str("[*,*]")
    assert ppn.group == ANY_GROUP
    assert ppn.user == ANY_USER


def test_ppn_from_word_normal_case():
    ppn = PPN.from_word(0x7B2D)  # 0x7B = 123, 0x2D = 45
    assert ppn.group == 123
    assert ppn.user == 45


def test_ppn_to_wide_str_normal_case():
    ppn = PPN(123, 45)
    assert ppn.to_wide_str() == "[123,45 ]"


def test_ppn_to_wide_str_any_group():
    ppn = PPN(ANY_GROUP, 45)
    assert ppn.to_wide_str() == "[  *,45 ]"


def test_ppn_to_wide_str_any_user():
    ppn = PPN(123, ANY_USER)
    assert ppn.to_wide_str() == "[123,*  ]"


def test_ppn_to_wide_str_any_group_and_user():
    ppn = PPN(ANY_GROUP, ANY_USER)
    assert ppn.to_wide_str() == "[  *,*  ]"


def test_ppn_str_normal_case():
    ppn = PPN(123, 45)
    assert str(ppn) == "[123,45]"


def test_ppn_str_any_group():
    ppn = PPN(ANY_GROUP, 45)
    assert str(ppn) == "[*,45]"


def test_ppn_str_any_user():
    ppn = PPN(123, ANY_USER)
    assert str(ppn) == "[123,*]"


def test_ppn_str_any_group_and_user():
    ppn = PPN(ANY_GROUP, ANY_USER)
    assert str(ppn) == "[*,*]"



# Copyright (C) 2014 Andrea Bonomi <andrea.bonomi@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import errno
import functools
import math
import operator
import os
import struct
import sys
import typing as t
from datetime import date, datetime, timedelta

from ..abstract import AbstractBlockFilesystem, AbstractDirectoryEntry, AbstractFile
from ..commons import (
    ASCII,
    BLOCK_SIZE,
    IMAGE,
    READ_FILE_FULL,
    bytes_to_word,
    dump_struct,
    filename_match,
    word_to_bytes,
)
from ..device.abstract import AbstractDevice
from ..unix.commons import unix_join, unix_split

__all__ = [
    "DGDOSFile",
    "DGDOSFilesystem",
    "rdos_canonical_filename",
    "rdos_get_file_type_id",
]

DISK_ID_BLOCK = 3  # Disk information block
SYS_DR = "SYS.DR"  # System Directory file
SYS_DR_BLOCK = 6  # First index block of SYS.DR (random file)
MAP_DR = "MAP.DR"  # Disk map file
MAP_DR_BLOCK = 15  # First block or MAP.DR (contiguous file)
SCPPA = 6  # Primary partition base address

UFD_ENTRY_FORMAT = "<10s2s12H"  # User File Descriptor Entry
UFD_ENTRY_LEN = struct.calcsize(UFD_ENTRY_FORMAT)
assert UFD_ENTRY_LEN == 36
UFD_LINK_ENTRY_FORMAT = "<10s2sH 10s10s2s"  # User File Descriptor Link Entry
UFD_LINK_ENTRY_LEN = struct.calcsize(UFD_LINK_ENTRY_FORMAT)
assert UFD_LINK_ENTRY_LEN == 36

START_DATE = datetime(1967, 12, 31)  # DGDOS date start

FILE_NAME_LENGTH = 10  # Maximum file name length
FILE_EXTENSION_LENGTH = 2  # Maximum file extension length

INDEX_ENTRIES = BLOCK_SIZE // 2 - 1  # Number of block addresses in an index block
SEQENTIAL_BLOCK_SIZE = BLOCK_SIZE - 2  # Bytes per block for sequential files
SEQENTIAL_BLOCK_SIZE_LARGE_DISK = BLOCK_SIZE - 4  # Bytes per block for sequential files on large disks

# Files attributes/characteristics
# Pag 61
# https://bitsavers.trailing-edge.com/pdf/dg/software/rdos/093-400027-00_RDOS_SystemReference_Oct83.pdf

ATWP = 2**0  #   Write protected - the file cannot be written
ATPER = 2**1  #  Permanent file - the file cannot be deleted or renamed
ATRAN = 2**2  #  Random file
ATCON = 2**3  #  Contiguous file
ATUS2 = 2**5  #  User defined 2
ATUS1 = 2**6  #  User defined 1
ATNRS = 2**8  #  No resolution allowed - the file cannot be linked
ATRES = 2**9  #  Link resolution file
ATDIR = 2**10  # Directory
ATPAR = 2**11  # Disk partition
ATLNK = 2**12  # Link entry
ATSAV = 2**13  # Save file (core image)
ATCHA = 2**14  # Attribute-protected file - the attributes cannot be changed
ATRP = 2**15  #  Read-protected file - the file cannot be read

# Disk ID Block - Disk characteristics

CHDOBL = 2**15  # Disk requires double addressing?
CHTOPL = 2**14  # Top-loader (dual-platter disk subsystem)

# File types

RANDOM_FILE_TYPE = 0  # Random file (indexed file)
CONTIGUOUS_FILE_TYPE = 1  # Contiguous file
SEQUENTIAL_FILE_TYPE = 2  # Sequential file (linked file)

FILE_TYPES = {
    RANDOM_FILE_TYPE: "RANDOM",
    CONTIGUOUS_FILE_TYPE: "CONTIGUOUS",
    SEQUENTIAL_FILE_TYPE: "SEQUENTIAL",
}


def rdos_to_date(date: int, time: int = 0) -> t.Optional[datetime]:
    """
    Convert RDOS date and time to datetime
    """
    if date == 0:
        return None
    return START_DATE + timedelta(
        days=date,
        hours=time >> 8,
        minutes=time % 256,
    )


def date_to_rdos(dt: t.Optional[datetime]) -> t.Tuple[int, int]:
    """
    Convert datetime to RDOS date and time
    """
    if dt is None:
        return 0, 0
    delta = dt - START_DATE
    days = delta.days
    hours = dt.hour
    minutes = dt.minute
    return days, (hours << 8) + minutes


def format_attr(attr: int, long: bool = False) -> str:
    if long:
        if (attr & ATDIR) != 0:  # Directory
            t = "DIR "
        elif (attr & ATPAR) != 0:  # Partition (RDOS only)
            t = "PART"
        elif (attr & ATLNK) != 0:  # Link entry
            t = "LINK"
        elif (attr & ATRAN) != 0:  # Random organized file
            t = "RAND"
        elif (attr & ATCON) != 0:  # Contiguous organized file
            t = "CONT"
        else:  # Sequential organized file
            t = "SEQ "
        return f"[{t}] " + "".join(
            [
                (attr & ATRP) and "R" or "",  #  read protected
                (attr & ATCHA) and "A" or "",  # attribute protected (attributes cannot be changed)
                (attr & ATPER) and "P" or "",  # permanent file (cannot be deleted or renamed)
                (attr & ATWP) and "W" or "",  #  write protected
                (attr & ATSAV) and "S" or "",  # save file (core image)
                (attr & ATUS2) and "&" or "",  # second user defined attribute
                (attr & ATUS1) and "?" or "",  # first user defined attribute
                (attr & ATNRS) and "N" or "",  # no resolution allowed (cannot be linked)
                (attr & ATRES) and "-" or "",  # link resolution file (temporary)
                # I ??? accessible by direct I/O only
            ]
        )

    return "".join(
        [
            (attr & ATRP) and "R" or "",  #  read protected
            (attr & ATCHA) and "A" or "",  # attribute protected (attributes cannot be changed)
            (attr & ATPER) and "P" or "",  # permanent file (cannot be deleted or renamed)
            (attr & ATWP) and "W" or "",  #  write protected
            (attr & ATSAV) and "S" or "",  # save file (core image)
            (attr & ATRAN) and "D" or "",  # random organized file
            (attr & ATCON) and "C" or "",  # contiguous organized file
            (attr & ATUS2) and "&" or "",  # second user defined attribute
            (attr & ATUS1) and "?" or "",  # first user defined attribute
            (attr & ATNRS) and "N" or "",  # no resolution allowed (cannot be linked)
            (attr & ATRES) and "-" or "",  # link resolution file (temporary)
            (attr & ATDIR) and "Y" or "",  # directory
            (attr & ATPAR) and "T" or "",  # partition
            (attr & ATLNK) and "L" or "",  # link entry
            # I ??? accessible by direct I/O only
        ]
    )


def swap_bytes(data: bytes) -> bytes:
    """
    Swap the bytes in a byte array
    """
    length = len(data)
    word_length = length // 2
    words = struct.unpack(f">{word_length}H", data[: word_length * 2])
    result = struct.pack(f"<{word_length}H", *words)
    # Append last byte if length is odd
    if length % 2:
        result += data[-1:]
    return result


def rdos_get_file_type_id(file_type: t.Optional[str], default: int = RANDOM_FILE_TYPE) -> int:
    """
    Get the file type id from a string
    """
    if not file_type:
        return default
    file_type = file_type.upper()
    for file_id, file_str in FILE_TYPES.items():
        if file_str == file_type:
            return file_id
    raise Exception("?KMON-F-Invalid file type specified with option")


def rdos_join(a: str, *p: str) -> str:
    """
    Join two or more pathname components
    """
    path = a
    if not p:
        path[:0] + "/"
    for b in p:
        if b.startswith("/"):
            path = b
        elif not path or path.endswith("/"):
            path += b
        else:
            path += "/" + b
    return path


def rdos_normpath(path: str, pwd: str) -> str:
    """
    Normalize path
    """
    if not path:
        return path
    if not path.startswith("/"):
        path = rdos_join(pwd, path)
    parts: t.List[str] = []
    for part in path.split("/"):
        if not part:
            pass
        elif part == ".":
            pass
        elif part == "..":
            if parts:
                parts.pop()
        else:
            part = rdos_canonical_filename(part)
            parts.append(part)
    return "/" + "/".join(parts)


def rdos_canonical_filename(basename: t.Optional[str], wildcard: bool = False) -> str:
    """
    Generate the canonical RDOS name
    """

    def filter_fn(s: str) -> bool:
        if wildcard and s == "*":
            return True
        return str.isalnum(s) or s == "$"

    if not basename:
        return ""

    try:
        filename, extension = basename.split(".", 1)
    except Exception:
        filename = basename
        extension = ""

    filename = "".join(filter(filter_fn, filename.upper()))[:FILE_NAME_LENGTH]
    extension = "".join(filter(filter_fn, extension.upper()))[:FILE_EXTENSION_LENGTH]
    if extension:
        return f"{filename}.{extension}"
    else:
        return filename


def bytes_to_ascii(val: bytes) -> str:
    """
    Convert bytes to ascii string
    """
    return swap_bytes(val).decode("ascii", errors="ignore").strip("\0")


def ascii_to_bytes(val: str, length: int) -> bytes:
    """
    Convert ascii string to bytes, padding with null bytes
    """
    return swap_bytes(val.encode("ascii").ljust(length, b"\0")[:length])


def filename_hash(filename: str, extension: str, frame_size: int = 0xFFFF) -> int:
    tmp = ascii_to_bytes(filename.upper(), FILE_NAME_LENGTH)
    t0 = sum([x * (0o400 if i % 2 == 1 else 1) for i, x in enumerate(tmp)])
    tmp = ascii_to_bytes(extension.upper(), FILE_EXTENSION_LENGTH)
    t1 = sum([x * (0o400 if i % 2 == 1 else 1) for i, x in enumerate(tmp)])
    return (t0 + t1) % 0xFFFF % frame_size


def words_dump(data: t.List[int], words_per_line: int = 8) -> None:
    """
    Display words in hexadecimal
    """
    is_zero = False
    for i in range(0, len(data), words_per_line):
        line = data[i : i + words_per_line]
        if not any(line):
            if not is_zero:
                sys.stdout.write("***\n")
            is_zero = True
            continue
        is_zero = False
        hex_str = "   ".join([f"{x:04X}" for x in line])
        tmp: t.List[int] = functools.reduce(operator.iconcat, [(x % 256, x >> 8) for x in line], [])
        ascii_str = "".join([chr(x) if 32 <= x <= 126 else "." for x in tmp])
        sys.stdout.write(f"{i:>6X}   {hex_str.ljust(3 * words_per_line)} {ascii_str}\n")


class DGDOSBitmap:

    fs: "DGDOSFilesystem"
    blocks: t.List[int]
    bitmaps: t.List[int]
    num_of_words: int

    def __init__(self, fs: "DGDOSFilesystem"):
        self.fs = fs

    @classmethod
    def read(cls, fs: "DGDOSFilesystem", parent_dir: t.Optional["UserFileDescriptor"] = None) -> "DGDOSBitmap":
        """
        Read the bitmap blocks
        """
        self = DGDOSBitmap(fs)
        entry = self.fs.get_ufd(parent_dir, MAP_DR)
        if not entry:
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), MAP_DR)
        self.num_of_words = entry.get_size() // 2
        self.blocks = list(entry.blocks())
        self.bitmaps = []
        for block in self.blocks:
            words = fs.read_16bit_words_block(block)
            if not words:
                raise OSError(errno.EIO, f"Failed to read block {block}")
            self.bitmaps.extend(words)
        self.bitmaps = self.bitmaps[: self.num_of_words]
        return self

    def write(self) -> None:
        """
        Write the bitmap blocks to the disk
        """
        for i, block in enumerate(self.blocks):
            words = self.bitmaps[i * BLOCK_SIZE // 2 : (i + 1) * BLOCK_SIZE // 2]
            if len(words) != BLOCK_SIZE // 2:
                words += [0] * (BLOCK_SIZE // 2 - len(words))
            self.fs.write_16bit_words(words, block)

    @property
    def total_bits(self) -> int:
        """
        Return the bitmap length in bit
        """
        return len(self.bitmaps) * 16

    def is_free(self, block_number: int) -> bool:
        """
        Check if a block is free
        """
        if block_number < SCPPA:
            return False
        bit_index = block_number - SCPPA
        int_index = bit_index // 16
        bit_position = bit_index % 16
        bit_value = self.bitmaps[int_index]
        return (bit_value & (1 << bit_position)) == 0

    def set_free(self, block_number: int) -> None:
        """
        Mark a block as free
        """
        if block_number < SCPPA:
            return
        bit_index = block_number - SCPPA
        int_index = bit_index // 16
        bit_position = bit_index % 16
        self.bitmaps[int_index] &= ~(1 << bit_position)

    def set_used(self, block_number: int) -> None:
        """
        Allocate a block
        """
        if block_number < SCPPA:
            return
        bit_index = block_number - SCPPA
        int_index = bit_index // 16
        bit_position = bit_index % 16
        self.bitmaps[int_index] |= 1 << bit_position

    def max_contiguous_blocks(self) -> int:
        """
        Find the maximum number of contiguous blocks
        """
        current_size = 0
        max_size = 0
        for block in range(SCPPA, self.total_bits + SCPPA):
            if self.is_free(block):
                current_size += 1
                max_size = max(max_size, current_size)
            else:
                current_size = 0
        return max_size

    def find_contiguous_blocks(self, size: int) -> int:
        """
        Find contiguous blocks, return the first block number
        """
        current_size = 0
        start_index = -1
        for block in range(SCPPA, self.total_bits + SCPPA):
            if self.is_free(block):
                if current_size == 0:
                    start_index = block
                current_size += 1
                if current_size == size:
                    return start_index
            else:
                current_size = 0
        raise OSError(errno.ENOSPC, os.strerror(errno.ENOSPC))

    def allocate(self, size: int, contiguous: bool = False) -> t.List[int]:
        """
        Allocate contiguous or sparse blocks
        """
        blocks = []
        if contiguous and size != 1:
            start_block = self.find_contiguous_blocks(size)
            for block in range(start_block, start_block + size):
                self.set_used(block)
                blocks.append(block)

        else:
            for block in range(SCPPA, self.total_bits + SCPPA):
                if self.is_free(block):
                    self.set_used(block)
                    blocks.append(block)
                if len(blocks) == size:
                    break
            if len(blocks) < size:
                raise OSError(errno.ENOSPC, os.strerror(errno.ENOSPC))
        return blocks

    def used(self) -> int:
        """
        Count the number of used blocks
        """
        used = 0
        for block in self.bitmaps:
            used += block.bit_count()
        return used

    def free(self) -> int:
        """
        Count the number of free blocks
        """
        return len(self.bitmaps) * 16 - self.used()

    def __str__(self) -> str:
        free = self.free()
        used = self.used()
        max_contiguous_blocks = self.max_contiguous_blocks()
        return f"LEFT: {free:<6} USED: {used:<6} MAX. CONTIGUOUS: {max_contiguous_blocks:6}"


class DGDOSFile(AbstractFile):
    entry: "UserFileDescriptor"
    closed: bool

    def __init__(self, entry: "UserFileDescriptor", file_mode: t.Optional[str] = None):
        self.entry = entry
        self.closed = False
        self.file_mode = file_mode or IMAGE

    def read_16bit_words_block(
        self,
        block_number: int,
        number_of_blocks: int = 1,
    ) -> t.List[int]:
        """
        Read 16bit words from file
        """
        buffer = self.read_block(block_number, number_of_blocks)
        if len(buffer) < BLOCK_SIZE:
            raise OSError(errno.EIO, os.strerror(errno.EIO))
        return list(struct.unpack("<256H", buffer))

    def read_block(
        self,
        block_number: int,
        number_of_blocks: int = 1,
    ) -> bytes:
        """
        Read block(s) of data from the file
        """
        if number_of_blocks == READ_FILE_FULL:
            number_of_blocks = self.entry.get_length()
        if (
            self.closed
            or block_number < 0
            or number_of_blocks < 0
            or block_number + number_of_blocks > self.entry.get_length()
        ):
            raise OSError(errno.EIO, os.strerror(errno.EIO))
        data = bytearray()
        # Get the blocks to be read
        blocks = list(self.entry.blocks())[block_number : block_number + number_of_blocks]
        # Read the blocks
        for disk_block_number in blocks:
            buffer = swap_bytes(self.entry.fs.read_block(disk_block_number))
            data.extend(buffer)
        # Convert to ASCII if needed
        if self.file_mode == ASCII:
            return bytes([0x0A if x == 0x0D else x for x in data])
        else:
            return bytes(data)

    def write_16bit_words(
        self,
        words: t.List[int],
        block_number: int,
        number_of_blocks: int = 1,
    ) -> None:
        """
        Write 16bit words to the file
        """
        buffer = struct.pack("<256H", *words)
        self.write_block(buffer, block_number, number_of_blocks)

    def write_block(
        self,
        buffer: bytes,
        block_number: int,
        number_of_blocks: int = 1,
    ) -> None:
        """
        Write block(s) of data to the file
        """
        if (
            self.closed
            or block_number < 0
            or number_of_blocks < 0
            or block_number + number_of_blocks > self.entry.get_length()
        ):
            raise OSError(errno.EIO, os.strerror(errno.EIO))
        # Convert to ASCII if needed
        if self.file_mode == ASCII:
            buffer = bytes([0x0D if x == 0x0A else x for x in buffer])
        block_size = self.get_block_size()
        # Get the blocks to be written
        blocks = list(self.entry.blocks())[block_number : block_number + number_of_blocks]
        # Write the blocks
        for i, disk_block_number in enumerate(blocks):
            data = swap_bytes(buffer[i * block_size : (i + 1) * block_size])
            self.entry.fs.write_block(data, disk_block_number)

    def get_length(self) -> int:
        """
        Get the length in blocks
        """
        return self.entry.get_length()

    def get_size(self) -> int:
        """
        Get file size in bytes
        """
        return self.entry.get_size()

    def get_block_size(self) -> int:
        """
        Get file block size in bytes
        """
        return self.entry.get_block_size()

    def close(self) -> None:
        """
        Close the file
        """
        self.closed = True

    def __str__(self) -> str:
        return self.entry.fullname


class UserFileDescriptor(AbstractDirectoryEntry):
    """
    User File Descriptor

    Word
        +-------------------------------------+
     0  |               File                  |
     4  |               name                  |
        +-------------------------------------+
     5  |            Extension                |
        +-------------------------------------+
     6  |   Attributes and characteristics    |
        +-------------------------------------+
     7  |       Link access attributes        |
        +-------------------------------------+
     8  |    Number of last block in file     |
        +-------------------------------------+
     9  |        Bytes in last block          |
        +-------------------------------------+
    10  |   First Block / Index Block number  |
        +-------------------------------------+
    11  |       Last access year/day          |
        +-------------------------------------+
    12  |    Last modification year/day       |
        +-------------------------------------+
    13  |   Hour and minute of last mod.      |
        +-------------------------------------+
    14  |        Variable information         |
    15  |                                     |
        +-------------------------------------+
    16  |           Use count                 |
        +-------------------------------------+
    17  |       Device code DCT link          |
        +-------------------------------------+

    Device code DCT link:
    - bits 10—16 - device code of device that holds file
    - bits 0—2   - for large disks, contain the high order of the disk address.

    Pag 25
    https://bitsavers.trailing-edge.com/pdf/dg/software/rdos/093-400027-00_RDOS_SystemReference_Oct83.pdf
    """

    sys_dir_block: "SystemDirectoryBlock"
    parent: t.Optional["UserFileDescriptor"]  # Parent directory/partition
    filename: str = ""
    extension: str = ""
    attributes: int = 0
    link_access_attributes: int = 0
    number_of_last_block: int = 0
    bytes_in_last_block: int = 0
    address: int = 0  # first block address for a sequential/contiguous file, index address for a random file
    last_access_date: int = 0  # days since 1967-12-31
    last_modification_date: int = 0  # days since 1967-12-31
    last_modification_time: int = 0  # hour (high byte) and minute (low byte)
    var1: int = 0
    var2: int = 0
    use_count: int = 0
    device_code: int = 0
    target: str = ""  # link target

    def __init__(self, sys_dir_block: "SystemDirectoryBlock", parent: t.Optional["UserFileDescriptor"] = None):
        self.sys_dir_block = sys_dir_block
        self.parent = parent

    @classmethod
    def read(
        cls,
        sys_dir_block: "SystemDirectoryBlock",
        parent: t.Optional["UserFileDescriptor"],
        buffer: bytes,
        position: int,
    ) -> "UserFileDescriptor":  # DOS Course Handouts, Pag 14
        # https://bitsavers.trailing-edge.com/pdf/dg/software/rdos/093-400027-00_RDOS_SystemReference_Oct83.pdf
        self = UserFileDescriptor(sys_dir_block, parent)
        (
            filename,
            extension,
            self.attributes,  #              1 word  Attributes and characteristics
            self.link_access_attributes,  #  1 word  Link access attributes
            self.number_of_last_block,  #    1 word  Number of last block in file
            self.bytes_in_last_block,  #     1 word  Byte count in last block
            self.address,  #                 1 byte  Address
            self.last_access_date,  #        1 byte  Year and day of last access
            self.last_modification_date,  #  1 word  Year and day of last modification
            self.last_modification_time,  #  1 word  Hour and minute of last modification
            self.var1,  #                    1 word  UFD variable information
            self.var2,  #                    1 word  UFD variable information
            self.use_count,  #               1 word  Use count
            self.device_code,  #             1 word  Device code DCT link
        ) = struct.unpack_from(UFD_ENTRY_FORMAT, buffer, position)
        if filename[0:2] == b'\0\0':  # deleted file/empty entry
            self.filename = ""
            self.extension = ""
        else:
            self.filename = bytes_to_ascii(filename)
            self.extension = bytes_to_ascii(extension)
        if self.is_link:
            (
                _,  # filename
                _,  # extension
                _,  # attributes,
                link_dir,
                link_name,
                link_ext,
            ) = struct.unpack_from(UFD_LINK_ENTRY_FORMAT, buffer, position)
            link_dir = bytes_to_ascii(link_dir)
            link_name = bytes_to_ascii(link_name)
            link_ext = bytes_to_ascii(link_ext)
            if link_dir:
                self.target = f"{link_dir}:{link_name}.{link_ext}"
            else:
                self.target = f"{link_name}.{link_ext}"
            self.number_of_last_block = 0
            self.bytes_in_last_block = 0
            self.last_modification_date = 0
            self.last_modification_time = 0
            self.last_access_date = 0
            self.var1 = 0
            self.var2 = 0
            self.use_count = 0
            self.device_code = 0
            pass
        return self

    @classmethod
    def create(
        cls,
        fs: "DGDOSFilesystem",
        parent: t.Optional["UserFileDescriptor"],
        filename: str,
        length: int,  # Length in blocks
        bitmap: "DGDOSBitmap",
        creation_date: t.Optional[t.Union[date, datetime]] = None,  # optional creation date
        file_type: t.Optional[str] = None,  # optional file type
        length_bytes: t.Optional[int] = None,  # optional length int bytes
    ) -> "UserFileDescriptor":
        """
        Create a new regular file
        """
        raw_file_type = rdos_get_file_type_id(file_type)
        if raw_file_type == RANDOM_FILE_TYPE:
            attributes = ATRAN
            index_blocks = (length - 1) // INDEX_ENTRIES + 1
            blocks_used = length + index_blocks
        elif raw_file_type == CONTIGUOUS_FILE_TYPE:
            attributes = ATCON
            blocks_used = length
        else:  # SEQUENTIAL_FILE_TYPE:
            attributes = 0
            blocks_used = length
        # Check the filename
        filename = rdos_canonical_filename(filename)
        if filename in [SYS_DR, MAP_DR]:
            raise OSError(errno.EEXIST, os.strerror(errno.EEXIST), filename)
        try:
            filename, extension = filename.split(".", 1)
        except Exception:
            extension = ""
        # Check free space and allocate blocks
        if raw_file_type == CONTIGUOUS_FILE_TYPE:
            allocated_blocks = bitmap.allocate(blocks_used, contiguous=True)
        else:
            if bitmap.free() < blocks_used:
                raise OSError(errno.ENOSPC, os.strerror(errno.ENOSPC))
            allocated_blocks = bitmap.allocate(blocks_used)
        # Get a free System Directory entry
        system_directory = SystemDirectory(fs, parent)
        entry = system_directory.get_free_entry(hash=filename_hash(filename, extension, fs.frame_size))
        if entry is None:
            raise OSError(errno.ENOSPC, os.strerror(errno.ENOSPC), "No free entries in SYS.DR")
        # Create the entry
        entry.filename = filename
        entry.extension = extension
        entry.attributes = attributes
        entry.address = allocated_blocks[0]
        entry.number_of_last_block = length - 1
        if length_bytes is not None:
            entry.bytes_in_last_block = length_bytes % BLOCK_SIZE
            if entry.bytes_in_last_block == 0:
                entry.bytes_in_last_block = BLOCK_SIZE
        else:
            entry.bytes_in_last_block = BLOCK_SIZE
        if creation_date is None:
            creation_date = datetime.now()
        entry.last_modification_date, entry.last_modification_time = date_to_rdos(creation_date)  # type: ignore
        entry.last_access_date = entry.last_modification_date
        # TODO
        entry.device_code = 27  # TODO
        # Block allocation
        if entry.is_random:  # Random file
            index_blocks = (length - 1) // INDEX_ENTRIES + 1  # Number of index blocks
            for i in range(index_blocks):
                words = allocated_blocks[index_blocks + i * INDEX_ENTRIES : index_blocks + (i + 1) * INDEX_ENTRIES]
                words = words + [0] * (INDEX_ENTRIES - len(words))
                if i < index_blocks - 1:
                    words.append(allocated_blocks[i + 1])  # Link to next index block
                else:
                    words.append(0)  # Last block
                index_block_number = allocated_blocks[i]
                fs.write_16bit_words(words, index_block_number)
        elif entry.is_sequential:  # Sequential file
            for i, block_number in enumerate(allocated_blocks):
                words = [0] * (BLOCK_SIZE // 2 - 1)
                if i < len(allocated_blocks) - 1:
                    words.append(allocated_blocks[i + 1])
                else:
                    words.append(0)  # Last block
                fs.write_16bit_words(words, block_number)
        # Write the entry
        entry.sys_dir_block.number_of_files += 1
        entry.sys_dir_block.write()
        return entry

    def write_buffer(self, buffer: bytearray, position: int) -> None:
        """
        Write the UFD entry to the buffer
        """
        filename = ascii_to_bytes(self.filename, FILE_NAME_LENGTH)
        extension = ascii_to_bytes(self.extension, FILE_EXTENSION_LENGTH)
        if self.is_link:
            try:
                link_dir, link_name = self.target.split(":", 1)
            except Exception:
                link_dir = ""
                link_name = self.target
            try:
                link_name, link_ext = link_name.split(".", 1)
            except Exception:
                link_ext = ""
            buffer[position : position + UFD_LINK_ENTRY_LEN] = struct.pack(
                UFD_LINK_ENTRY_FORMAT,
                filename,
                extension,
                self.attributes,
                ascii_to_bytes(link_dir, FILE_NAME_LENGTH),
                ascii_to_bytes(link_name, FILE_NAME_LENGTH),
                ascii_to_bytes(link_ext, FILE_EXTENSION_LENGTH),
            )
        else:
            buffer[position : position + UFD_ENTRY_LEN] = struct.pack(
                UFD_ENTRY_FORMAT,
                filename,
                extension,
                self.attributes,
                self.link_access_attributes,
                self.number_of_last_block,
                self.bytes_in_last_block,
                self.address,
                self.last_access_date,
                self.last_modification_date,
                self.last_modification_time,
                self.var1,
                self.var2,
                self.use_count,
                self.device_code,
            )

    @property
    def fs(self) -> "DGDOSFilesystem":
        """
        Get the filesystem
        """
        return self.sys_dir_block.fs

    @property
    def is_random(self) -> bool:
        """
        Check if the file is random organized
        """
        return self.attributes & ATRAN != 0

    @property
    def is_contiguous(self) -> bool:
        """
        Check if the file is contiguous organized
        """
        return self.attributes & ATCON != 0

    @property
    def is_sequential(self) -> bool:
        """
        Check if the file is sequential organized
        """
        return not self.is_random and not self.is_contiguous and not self.is_link

    @property
    def is_link(self) -> bool:
        """
        Check if the file is a link
        """
        return self.attributes & ATLNK != 0

    @property
    def is_directory(self) -> bool:
        """
        Check if the file is a directory
        """
        return self.attributes & ATDIR != 0

    @property
    def is_partition(self) -> bool:
        """
        Check if the file is a partition
        """
        return self.attributes & ATPAR != 0

    @property
    def is_empty(self) -> bool:
        return self.filename == "" and self.extension == ""

    @property
    def fullname(self) -> str:
        if self.parent:
            return rdos_join(self.parent.fullname, self.basename)
        else:
            return self.basename

    @property
    def basename(self) -> str:
        return f"{self.filename}.{self.extension}"

    @property
    def last_access(self) -> t.Optional[date]:
        """
        Last access date
        """
        return rdos_to_date(self.last_access_date)

    @property
    def creation_date(self) -> t.Optional[date]:
        """
        Last modification date
        """
        return rdos_to_date(self.last_modification_date, self.last_modification_time)

    def get_length(self) -> int:
        """
        Get the length in blocks
        """
        if self.bytes_in_last_block == 0:
            return self.number_of_last_block
        else:
            return self.number_of_last_block + 1

    def get_size(self) -> int:
        """
        Get file size in bytes
        """
        return self.number_of_last_block * BLOCK_SIZE + self.bytes_in_last_block

    def get_block_size(self) -> int:
        """
        Get file block size in bytes
        """
        if self.is_sequential:  # Sequential file block size
            if self.fs.double_addressing:
                return SEQENTIAL_BLOCK_SIZE_LARGE_DISK
            else:
                return SEQENTIAL_BLOCK_SIZE
        else:
            return BLOCK_SIZE

    def filename_hash(self) -> int:
        if self.is_empty:
            return 0
        return filename_hash(self.filename, self.extension, self.fs.frame_size)

    def delete(self) -> bool:
        """
        Delete the file
        """
        # If this is a directory, delete the directory entries
        if self.is_directory:
            for subentry in self.fs.read_dir_entries(self):
                if subentry.basename != MAP_DR and self.address != subentry.address:
                    subentry.delete()
        # Deallocate the blocks
        bitmap = self.fs.read_bitmap(self.parent)
        for block_number in self.blocks(include_indexes=True):
            bitmap.set_free(block_number)
        bitmap.write()
        # Deallocate the entry
        self.filename = ""
        self.extension = ""
        self.attributes = 0
        self.link_access_attributes = 0
        self.number_of_last_block = 0
        self.address = 0
        self.last_access_date = 0
        self.last_modification_date = 0
        self.last_modification_time = 0
        self.var1 = 0
        self.var2 = 0
        self.use_count = 0
        self.device_code = 0
        # Decrement the number of files in the System Directory Block
        if self.sys_dir_block.number_of_files > 0:
            self.sys_dir_block.number_of_files -= 1
        # Write the System Directory Block
        self.sys_dir_block.write()
        return True

    def write(self) -> bool:
        """
        Write the directory entry
        """
        self.sys_dir_block.write()
        return True

    def examine(self) -> str:
        if self.is_directory:  # Directory
            file_type = "Directory"
        elif self.is_partition:  # Partition
            file_type = "Partition"
        elif self.is_link:  # Link entry
            file_type = "Link"
        elif self.is_random:  # Random organized file
            file_type = "Random file"
        elif self.is_contiguous:  # Contiguous organized file
            file_type = "Contiguous file"
        else:  # Sequential organized file
            file_type = "Sequential file"
        if self.is_link:
            data: t.Dict[str, t.Any] = {
                "Filename": self.fullname,
                "File type": file_type,
                "Creation date": str(self.creation_date),
                "Target": self.target,
            }
        else:
            data = {
                "Filename": self.fullname,
                "File type": file_type,
                "Creation date": str(self.creation_date),
                "Last access": str(self.last_access),
                "Address": self.address,
                "Blocks ": self.get_length(),
                "File size": f"{self.get_size()} ({self.number_of_last_block} blocks + {self.bytes_in_last_block} bytes)",
                "Write protected": self.attributes & ATWP != 0,
                "Read protected": self.attributes & ATRP != 0,
                "Immutable attribs": self.attributes & ATCHA != 0,
                "Permanent": self.attributes & ATPER != 0,
                "Link attributes": format_attr(self.link_access_attributes),
                "Blocks": list(self.blocks()),
            }
        data["Filename hash"] = self.filename_hash()
        return dump_struct(data) + "\n"

    def open(self, file_mode: t.Optional[str] = None) -> DGDOSFile:
        """
        Open a file
        """
        return DGDOSFile(self, file_mode)

    def blocks(self, include_indexes: bool = False) -> t.Iterator[int]:
        if self.is_random:
            # Random file
            # Pag 22
            # https://bitsavers.trailing-edge.com/pdf/dg/software/rdos/093-400027-00_RDOS_SystemReference_Oct83.pdf
            index_block_number = self.address
            while index_block_number:
                if include_indexes:
                    yield index_block_number
                words = self.fs.read_16bit_words_block(index_block_number)
                index_block_number = words.pop()  # Link to the next file index block
                for block_address in words:
                    if block_address != 0:
                        yield block_address
        elif self.is_contiguous:
            # Contiguous file
            # Pag 23
            # https://bitsavers.trailing-edge.com/pdf/dg/software/rdos/093-400027-00_RDOS_SystemReference_Oct83.pdf
            length = self.get_length()
            for block_address in range(self.address, self.address + length):
                yield block_address
        else:
            # Sequential file
            # Pag 21
            # https://bitsavers.trailing-edge.com/pdf/dg/software/rdos/093-400027-00_RDOS_SystemReference_Oct83.pdf
            block_number = self.address
            while block_number:
                yield block_number
                words = self.fs.read_16bit_words_block(block_number)
                block_number = words[-1]
                if self.fs.double_addressing:
                    block_number += words[-2] << 16
                    # print(">>>", words[-2], words[-1], block_number, block_number)
                if block_number == 0:
                    break

    def __str__(self) -> str:
        attr = format_attr(self.attributes, long=True)
        if self.is_link:
            return f"{self.filename:>10s}.{self.extension:<2s} {attr:<12}  -> {self.target}"
        else:
            uftlkl = format_attr(self.link_access_attributes)
            if uftlkl:
                attr = f"{attr}/{uftlkl}"
            creation_date = self.creation_date.strftime("%m/%d/%y") if self.creation_date else ""
            return (
                f"{self.filename:>10s}.{self.extension:<2s} "
                f"{attr:<12} "
                f"{self.get_size():>10d}  "
                f"{creation_date:<8}  "
                f"{self.address:>6} {self.use_count:>5} {self.device_code:>5}"
            )

    def __repr__(self) -> str:
        return str(self)


class SystemDirectoryBlock:
    """
    System Directory Block

    Word
          +-------------------------------------+
    0     |          Number of files            |
          +-------------------------------------+
    1     |    User File Descriptor (UFD) #0    |
    17    | .                                   |
          +-------------------------------------+
    18    |    User File Descriptor (UFD) #1    |
    35    | .                                   |
          +-------------------------------------+
    36    |              UFDs ...               |
          | .                                   |
          +-------------------------------------+
    254   |     Max number UFD in this block    |
          +-------------------------------------+

    RDOS System Reference - Pag 24
    https://bitsavers.trailing-edge.com/pdf/dg/software/rdos/093-400027-00_RDOS_SystemReference_Oct83.pdf
    """

    sys_dir: "SystemDirectory"  # System Directory
    block_number: int = 0  # Block number
    number_of_files: int = 0  # Number of files in this block directory
    max_ufd: int = 0  # Max number of UFD in this block
    entries_list: t.List["UserFileDescriptor"] = []  # List of UFD entries

    def __init__(self, sys_dir: "SystemDirectory"):
        self.sys_dir = sys_dir

    @classmethod
    def read(cls, sys_dir: "SystemDirectory", block_number: int) -> "SystemDirectoryBlock":
        self = cls(sys_dir)
        self.block_number = block_number
        buffer = self.fs.read_block(block_number)
        self.number_of_files = bytes_to_word(buffer, 0)  # first word
        self.max_ufd = bytes_to_word(buffer[-2:], 0)  # last word
        self.entries_list = []
        for position in range(2, BLOCK_SIZE - 4 - UFD_ENTRY_LEN, UFD_ENTRY_LEN):
            ufd = UserFileDescriptor.read(self, sys_dir.dir_ufd, buffer, position)
            self.entries_list.append(ufd)
        return self

    def to_bytes(self) -> bytes:
        """
        Write the UFD entries to a buffer
        """
        buffer = bytearray(BLOCK_SIZE)
        # Update the number of files
        self.number_of_files = len([x for x in self.entries_list if not x.is_empty])
        buffer[0:2] = word_to_bytes(self.number_of_files)
        # Write the UFD entries
        for i, entry in enumerate(self.entries_list):
            entry.write_buffer(buffer, 2 + i * UFD_ENTRY_LEN)
        # Write the last word
        buffer[-2:] = word_to_bytes(self.max_ufd)
        return bytes(buffer)

    def write(self) -> None:
        """
        Write the System Directory Block to the disk
        """
        self.fs.write_block(self.to_bytes(), self.block_number)

    def iterdir(self) -> t.Iterator["UserFileDescriptor"]:
        """
        Iterate over directory entries
        """
        for entry in self.entries_list:
            if not entry.is_empty:
                yield entry

    @property
    def fs(self) -> "DGDOSFilesystem":
        """
        Get the filesystem
        """
        return self.sys_dir.fs

    def __str__(self) -> str:
        return f"Block {self.block_number} ({self.number_of_files} entries, {self.max_ufd})"


class SystemDirectory:
    """
    System Directory (SYS.DR)
    """

    fs: "DGDOSFilesystem"
    dir_ufd: t.Optional["UserFileDescriptor"] = None  # Directory entry

    def __init__(self, fs: "DGDOSFilesystem", dir_ufd: t.Optional["UserFileDescriptor"] = None):
        self.fs = fs
        self.dir_ufd = dir_ufd

    def blocks(self) -> t.Iterator[int]:
        """
        Get the blocks of the System Directory
        (SYS.DR is a random file)
        """
        index_block_number = self.dir_ufd.address if self.dir_ufd else SYS_DR_BLOCK
        while index_block_number:
            words = self.fs.read_16bit_words_block(index_block_number)
            index_block_number = words.pop()
            for block_number in words:
                if block_number != 0:
                    yield block_number

    def read_system_directory(self) -> t.Iterator[SystemDirectoryBlock]:
        """
        Read System Directory blocks
        """
        for block_number in self.blocks():
            yield SystemDirectoryBlock.read(self, block_number)

    def get_free_entry(self, hash: int) -> t.Optional["UserFileDescriptor"]:
        """
        Get a free System Directory entry
        """
        blocks = list(self.blocks())
        # TODO check
        for d in (0, -1):
            block_number = blocks[(hash + d) % len(blocks)]
            sys_dir_block = SystemDirectoryBlock.read(self, block_number)
            for entry in sys_dir_block.entries_list:
                if entry.is_empty:
                    return entry
        # for sys_dir_block in self.read_system_directory():
        #     for entry in sys_dir_block.entries_list:
        #         if entry.is_empty:
        #             return entry
        # TODO overlay
        return None


class DiskInformationBlock:
    """
    Disk Information Block

    Word
          +-------------------------------------+
    0     |          Revision number            |
          +-------------------------------------+
    1     |             Checksum                |
          +-------------------------------------+
    2     |         Number of heads             |
          +-------------------------------------+
    3     |       Number of sectors/track       |
          +-------------------------------------+
    4     |   Number of blocks/device high      |
          +-------------------------------------+
    5     |   Number of blocks/device low       |
          +-------------------------------------+
    6     |            Frame size               |
          +-------------------------------------+
    7     |           Characteristics           |
          +-------------------------------------+

    Frame size is the number of blocks initially allocated
    for the system directory.
    """

    fs: "DGDOSFilesystem"
    revision: int = 0
    checksum: int = 0
    heads: int = 0
    sectors: int = 0
    blocks: int = 0
    frame_size: int = 0
    characteristics: int = 0

    def __init__(self, fs: "DGDOSFilesystem"):
        self.fs = fs

    @classmethod
    def read(cls, fs: "DGDOSFilesystem") -> "DiskInformationBlock":
        self = cls(fs)
        (
            self.revision,
            self.checksum,
            self.heads,
            self.sectors,
            blocks_h,
            blocks_l,
            self.frame_size,
            self.characteristics,
        ) = self.fs.read_16bit_words_block(DISK_ID_BLOCK)[0:8]
        self.blocks = (blocks_h << 16) + blocks_l + SCPPA
        return self

    def is_double_addressing(self) -> bool:
        """
        Check if the disk requires double addressing
        """
        return self.characteristics & CHDOBL != 0

    def is_top_loader(self) -> bool:
        """
        Check if the disk is a top loader (dual-platter disk subsystem)
        """
        return self.characteristics & CHTOPL != 0

    def __str__(self) -> str:
        return (
            "\n*Disk Information Block\n"
            f"Revision:          {self.revision}\n"
            f"Checksum:          ${self.checksum:04x}\n"
            f"Heads:             {self.heads}\n"
            f"Sectors/track:     {self.sectors}\n"
            f"Blocks:            {self.blocks}\n"
            f"Frame size:        {self.frame_size}\n"
            f"Characteristics:   {self.characteristics:016b} (${self.characteristics:04x})\n"
            f"Double addressing: {self.is_double_addressing()}\n"
            f"Top loader:        {self.is_top_loader()}\n"
        )


class DGDOSFilesystem(AbstractBlockFilesystem):
    """
    DGDOS Filesystem


    Block
          +-------------------------------------+
    0     |            Bootstrap block          |
    1     |                                     |
          +-------------------------------------+
    2     |                                     |
          +-------------------------------------+
    3     |               Disk ID               |
          +-------------------------------------+
    4     |         Bad block pool index        |
    5     |                                     |
          +-------------------------------------+
    6     |    First index block of SYS.DR      |
          +-------------------------------------+
    7     |               Swap                  |
    14    |                                     |
          +-------------------------------------+
    15    |           MAP.DR blocks             |
    n     |                                     |
          +-------------------------------------+
    n+1   |            BOOTSYS.OL               |
    m     |                                     |
          +--------------------------------------

    RDOS System Reference - Pag 17
    https://bitsavers.trailing-edge.com/pdf/dg/software/rdos/093-400027-00_RDOS_SystemReference_Oct83.pdf
    """

    fs_name = "rdos"
    fs_description = "Data General Nova DOS/RDOS Filesystem"
    fs_platforms = ["nova"]

    double_addressing: bool = False  # Disk requires double addressing
    top_loader: bool = False  # Disk is a top loader
    frame_size: int = 1  # Frame size
    heads: int = 0  # Number of heads
    sectors_per_track: int = 0  # Number of sectors per track
    swap: bool = False  # Swap bytes
    pwd: str = "/"  # Current working directory

    def physical_block(self, logical_block: int) -> int:
        """
        Convert logical block number to physical block number
        """
        # TODO check if the condition is correct
        if self.heads == 0 or self.sectors_per_track == 0 or not self.top_loader:
            return logical_block
        else:
            h = self.heads * self.sectors_per_track
            return logical_block + (logical_block // h) * h

    def read_block(
        self,
        block_number: int,
        number_of_blocks: int = 1,
    ) -> bytes:
        block_number = self.physical_block(block_number)
        buffer = super().read_block(block_number, number_of_blocks)
        if self.swap:
            buffer = swap_bytes(buffer)
        return buffer

    def write_block(
        self,
        buffer: bytes,
        block_number: int,
        number_of_blocks: int = 1,
    ) -> None:
        block_number = self.physical_block(block_number)
        if self.swap:
            buffer = swap_bytes(buffer)
        super().write_block(buffer, block_number, number_of_blocks)

    def read_16bit_words_block(self, block_number: int) -> t.List[int]:
        """
        Read 256 16bit words from a block
        """
        buffer = self.read_block(block_number)
        if len(buffer) < BLOCK_SIZE:
            raise OSError(errno.EIO, os.strerror(errno.EIO))
        return list(struct.unpack("<256H", buffer))

    def write_16bit_words(self, words: t.List[int], block_number: int) -> None:
        """
        Write 256 16bit words to a block
        """
        buffer = struct.pack("<256H", *words)
        self.write_block(buffer, block_number)

    def check_map_dr(self) -> bool:
        """
        Check if the MAP.DR file is present
        """
        try:
            for tmp in self.read_dir_entries():
                if tmp.basename == MAP_DR:
                    return True
        except Exception:
            pass
        return False

    @classmethod
    def mount(
        cls,
        file_or_dev: t.Union["AbstractFile", "AbstractDevice"],
        strict: t.Union[bool, str] = True,
        **kwargs: t.Union[bool, str],
    ) -> "DGDOSFilesystem":
        self = cls(file_or_dev)
        # Read the Disk Information Block
        disk_id = DiskInformationBlock.read(self)
        if disk_id.revision > 16:
            self.swap = True
            disk_id = DiskInformationBlock.read(self)
        self.double_addressing = disk_id.is_double_addressing()
        self.top_loader = disk_id.is_top_loader()
        self.frame_size = disk_id.frame_size
        self.heads = disk_id.heads
        self.sectors_per_track = disk_id.sectors
        if strict:
            # Check if the file is a valid DOS/RDOS filesystem
            if not self.check_map_dr():
                raise OSError(errno.EIO, "MAP.DR not found")
        return self

    def read_dir_entries(
        self,
        dir_ufd: t.Optional["UserFileDescriptor"] = None,
    ) -> t.Iterator["UserFileDescriptor"]:
        """
        Read System Directory entries
        """
        system_directory = SystemDirectory(self, dir_ufd)
        for sys_dir_block in system_directory.read_system_directory():
            yield from sys_dir_block.iterdir()

    def read_bitmap(self, parent: t.Optional[UserFileDescriptor] = None) -> DGDOSBitmap:
        """
        Read the bitmap
        """
        if parent is None:
            parent = self.get_file_entry(self.pwd)
        return DGDOSBitmap.read(self, parent)

    def filter_entries_list(
        self,
        pattern: t.Optional[str],
        include_all: bool = False,
        expand: bool = True,
        wildcard: bool = True,
    ) -> t.Iterator["UserFileDescriptor"]:
        if pattern:
            pattern = pattern.upper()
        if not pattern and expand:
            pattern = "*"
        if pattern and pattern.startswith("/"):
            absolute_path = pattern
        else:
            absolute_path = unix_join(self.pwd, pattern or "")
        if self.isdir(absolute_path):
            if not expand:
                yield self.get_file_entry(absolute_path)
                return
            dirname = pattern
            pattern = "*"
        else:
            dirname, pattern = unix_split(absolute_path)
        if dirname == "/":  # Root directory
            dir_ufd: t.Optional[UserFileDescriptor] = None
        else:
            dir_ufd = self.get_file_entry(dirname)  # type: ignore
        for entry in self.read_dir_entries(dir_ufd):
            if filename_match(entry.basename, pattern, wildcard):
                yield entry

    def get_ufd(self, dir_ufd: t.Optional[UserFileDescriptor], basename: str) -> UserFileDescriptor:
        """
        Get User File Descriptor for a file in a directory
        """
        basename = rdos_canonical_filename(basename).rstrip(".")
        for entry in self.read_dir_entries(dir_ufd):
            if filename_match(entry.basename.rstrip("."), basename, wildcard=False):
                return entry
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), basename)

    @property
    def entries_list(self) -> t.Iterator["UserFileDescriptor"]:
        yield from self.read_dir_entries()

    def get_file_entry(self, fullname: str) -> UserFileDescriptor:
        """
        Get the directory entry for a file
        """
        fullname = unix_join(self.pwd, fullname) if not fullname.startswith("/") else fullname
        parts = [x for x in fullname.split("/") if x]
        if not parts:
            parts = [SYS_DR]  # Default to SYS.DR
        entry: t.Optional[UserFileDescriptor] = None
        for i, part in enumerate(parts):
            if entry is not None and not (entry.is_directory or entry.is_partition):  # is a directory?
                raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), fullname)
            entry = self.get_ufd(entry, part)
        if entry is None:
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), fullname)
        return entry

    def write_bytes(
        self,
        fullname: str,
        content: bytes,
        creation_date: t.Optional[date] = None,
        file_type: t.Optional[str] = None,
        file_mode: t.Optional[str] = None,
    ) -> None:
        """
        Write content to a file
        """
        raw_file_type = rdos_get_file_type_id(file_type)
        # Determine the number of blocks
        if raw_file_type == SEQUENTIAL_FILE_TYPE:
            block_size = SEQENTIAL_BLOCK_SIZE_LARGE_DISK if self.double_addressing else SEQENTIAL_BLOCK_SIZE
        else:
            block_size = BLOCK_SIZE
        number_of_blocks = int(math.ceil(len(content) * 1.0 / block_size))
        # Create the file
        entry = self.create_file(
            fullname=fullname,
            number_of_blocks=number_of_blocks,
            creation_date=creation_date,
            file_type=file_type,
            length_bytes=len(content),
        )
        # Write the content to the file
        if entry is not None:
            f = entry.open(file_mode)
            try:
                f.write(content)
            finally:
                f.close()

    def create_file(
        self,
        fullname: str,
        number_of_blocks: int,  # length in blocks
        creation_date: t.Optional[date] = None,  # optional creation date
        file_type: t.Optional[str] = None,
        length_bytes: t.Optional[int] = None,  # optional length in bytes
    ) -> t.Optional[UserFileDescriptor]:
        """
        Create a new file with a given length in number of blocks
        """
        fullname = rdos_normpath(fullname, self.pwd)
        dirname, filename = unix_split(fullname)
        # Delete the file if it already exists
        try:
            self.get_file_entry(fullname).delete()
        except FileNotFoundError:
            pass
        # Get parent directory
        try:
            parent: UserFileDescriptor = self.get_file_entry(dirname)
        except FileNotFoundError:
            raise NotADirectoryError(errno.ENOTDIR, os.strerror(errno.ENOTDIR), dirname)
        # Create the file
        bitmap = self.read_bitmap(parent)
        entry = UserFileDescriptor.create(
            fs=self,
            parent=parent,
            filename=filename,
            length=number_of_blocks,
            bitmap=bitmap,
            creation_date=creation_date,
            # access=access if access is not None else DEFAULT_ACCESS,
            file_type=file_type,
            # aux_type=aux_type if aux_type is not None else 0,
            length_bytes=length_bytes if length_bytes is not None else number_of_blocks * BLOCK_SIZE,
        )
        bitmap.write()
        return entry

    def isdir(self, fullname: str) -> bool:
        try:
            entry = self.get_file_entry(fullname)
            return entry.is_directory or entry.is_partition
        except FileNotFoundError:
            return False

    def dir(self, volume_id: str, pattern: t.Optional[str], options: t.Dict[str, bool]) -> None:
        for x in self.filter_entries_list(pattern, include_all=True, wildcard=True):
            if options.get("brief"):
                # For brief mode, print only the file name
                sys.stdout.write(f"{x.basename}\n")
            elif x.is_link:
                # Print link information
                # filename, target
                sys.stdout.write(f"{x.basename:<13s}             {x.target}\n")
            else:
                # Print file information
                # filename, byte length, attributes, last modification date, last access date, address, use count
                attr = format_attr(x.attributes)
                uftlkl = format_attr(x.link_access_attributes)
                if uftlkl:
                    attr = f"{attr}/{uftlkl}"
                creation_date = x.creation_date.strftime("%m/%d/%y %H:%M") if x.creation_date else ""
                access_date = x.last_access.strftime("%m/%d/%y") if x.last_access else ""
                sys.stdout.write(
                    f"{x.basename:<13s}{x.get_size():>10d}  {attr:<7} {creation_date:<14} {access_date:<8}  [{x.address:06o}] {x.use_count:>5}\n"
                )
        sys.stdout.write("\n")

    def examine(self, arg: t.Optional[str], options: t.Dict[str, t.Union[bool, str]]) -> None:
        if options.get("diskid"):
            # Display the filesystem information
            disk_id = DiskInformationBlock.read(self)
            sys.stdout.write(str(disk_id))
        elif options.get("free"):
            # Display the free space
            bitmap = self.read_bitmap()
            sys.stdout.write(f"{bitmap}\n")
        elif options.get("bitmap"):
            # Display the bitmap
            bitmap = self.read_bitmap()
            for i in range(0, bitmap.total_bits):
                sys.stdout.write(f"{i:>4} {'[ ]' if bitmap.is_free(i) else '[X]'}  ")
                if i % 16 == 15:
                    sys.stdout.write("\n")
        elif not arg:
            # Display the system directory
            sys.stdout.write("\n*System Directory\n")
            system_directory = SystemDirectory(self)
            for i, sys_dir_block in enumerate(system_directory.read_system_directory()):
                if options.get("full"):
                    sys.stdout.write(f"** #{i} {sys_dir_block}\n")
                for entry in sys_dir_block.entries_list:
                    if options.get("full") or not entry.is_empty:
                        sys.stdout.write(f"{entry} -- {entry.filename_hash() }\n")
                        # sys.stdout.write(f"{entry}\n")
        else:
            # Display the file information
            entry = self.get_file_entry(arg)  # type: ignore
            sys.stdout.write(entry.examine())

    def dump(self, fullname: t.Optional[str], start: t.Optional[int] = None, end: t.Optional[int] = None) -> None:
        """Dump the content of a file or a range of blocks"""
        if fullname:
            if start is None:
                start = 0
            if end is None:
                entry = self.get_file_entry(fullname)
                end = entry.get_length() - 1
            f: DGDOSFile = t.cast(DGDOSFile, self.open_file(fullname, file_mode=IMAGE))
            try:
                for block_number in range(start, end + 1):
                    data = f.read_16bit_words_block(block_number)
                    sys.stdout.write(f"\nBLOCK NUMBER   {block_number:08}\n")
                    words_dump(data)
            finally:
                f.close()
        else:
            if start is None:
                start = 0
            if end is None:
                if start == 0:
                    end = self.get_size() // BLOCK_SIZE
                else:
                    end = start
            for block_number in range(start, end + 1):
                data = self.read_16bit_words_block(block_number)
                sys.stdout.write(f"\nBLOCK NUMBER   {block_number:08}\n")
                words_dump(data)

    def get_size(self) -> int:
        """
        Get filesystem size in bytes
        """
        return self.dev.get_size()

    def chdir(self, fullname: str) -> bool:
        """
        Change the current directory
        """
        fullname = rdos_normpath(fullname, self.pwd)
        if not self.isdir(fullname):
            return False
        # Check if the directory is valid
        if any(x == SYS_DR for x in fullname.split("/")):
            return False
        self.pwd = fullname
        return True

    def get_pwd(self) -> str:
        """
        Get the current directory
        """
        return self.pwd

    def get_types(self) -> t.List[str]:
        """
        Get the list of the supported file types
        """
        return list(FILE_TYPES.values())
